@@ -1,11 +1,11 @@
 /****************************************************************************
-** $Id$
+** $Id: qt/qpainter_qws.cpp   3.1.2   edited Nov 7 16:55 $
 **
-** Implementation of QPainter class for FB
+** Implementation of QPainter class for Qt/Embedded
 **
 ** Created : 991026
 **
-** Copyright (C) 1992-2000 Trolltech AS.  All rights reserved.
+** Copyright (C) 1992-2002 Trolltech AS.  All rights reserved.
 **
 ** This file is part of the kernel module of the Qt GUI Toolkit.
 **
@@ -41,13 +41,11 @@
 #include "qtextcodec.h"
 #include "qmemorymanager_qws.h"
 #include "qwsdisplay_qws.h"
-#include <ctype.h>
-#include <stdlib.h>
 #include "qpaintdevicemetrics.h"
-//#include "qimagepaintdevice.h"
 #include "qgfx_qws.h"
+#include <string.h>
 
-/* paintevent magic to provide Windows semantics on FB
+/* paintevent magic to provide Windows semantics on Qt/E
  */
 static QRegion* paintEventClipRegion = 0;
 static QRegion* paintEventSaveRegion = 0;
@@ -82,6 +80,12 @@ void qt_clear_paintevent_clipping()
     paintEventDevice = 0;
 }
 
+class QPainterPrivate
+{
+public:
+    QPoint curPt;
+};
+
 
 /*****************************************************************************
   Trigonometric function for QPainter
@@ -167,10 +171,10 @@ inline double qcos( double a ) { return qsincos(a,TRUE); }
   QPainter member functions
  *****************************************************************************/
 
-const int TxNone      = 0;			// transformation codes
-const int TxTranslate = 1;			// also in qpainter.cpp
-const int TxScale     = 2;
-const int TxRotShear  = 3;
+static const int TxNone      = 0;		// transformation codes
+static const int TxTranslate = 1;		// also in qpainter.cpp
+static const int TxScale     = 2;
+static const int TxRotShear  = 3;
 
 
 void QPainter::initialize()
@@ -184,7 +188,7 @@ void QPainter::cleanup()
 
 void QPainter::destroy()
 {
-
+    delete d;
 }
 
 
@@ -217,7 +221,7 @@ void QPainter::redirect( QPaintDevice *pdev, QPaintDevice *replacement )
 
 void QPainter::init()
 {
-    d = 0;
+    d = new QPainterPrivate;
     flags = IsStartingUp;
     bg_col = white;				// default background color
     bg_mode = TransparentMode;			// default background mode
@@ -258,9 +262,11 @@ void QPainter::updateFont()
 {
     clearf(DirtyFont);
     if(testf(ExtDev)) {
+	if ( pfont ) delete pfont;
+	pfont = new QFont( cfont.d, pdev );
 	QPDevCmdParam param[1];
 	param[0].font = &cfont;
-	if ( !pdev->cmd( QPaintDevice::PdcSetFont, this, param ) )
+	if ( !pdev->cmd(QPaintDevice::PdcSetFont, this, param) )
 	    return;
     }
     if ( gfx )
@@ -273,7 +279,7 @@ void QPainter::updatePen()
     if(testf(ExtDev)) {
 	QPDevCmdParam param[1];
 	param[0].pen = &cpen;
-	if ( !pdev->cmd( QPaintDevice::PdcSetPen, this, param ) )
+	if ( !pdev->cmd(QPaintDevice::PdcSetPen, this, param) )
 	    return;
     }
     if ( gfx )
@@ -328,50 +334,48 @@ static uchar *pat_tbl[] = {
     dense6_pat, dense7_pat,
     hor_pat, ver_pat, cross_pat, bdiag_pat, fdiag_pat, dcross_pat };
 
-
-   if(testf(ExtDev)) {
-        QPDevCmdParam param[1];
-        param[0].brush = &cbrush;
-        if ( !pdev->cmd( QPaintDevice::PdcSetBrush, this, param ) )
+    if ( testf(ExtDev) ) {
+	QPDevCmdParam param[1];
+	param[0].brush = &cbrush;
+        if ( !pdev->cmd(QPaintDevice::PdcSetBrush, this, param) )
 	    return;
-   }
+    }
+    if ( !gfx )
+	return;
 
-   if ( !gfx )
-       return;
-
-   uchar * pat=0;
-   int bs=cbrush.style();
-   int d=0;
-   if( bs>=Dense1Pattern && bs <= DiagCrossPattern ) {
-       pat=pat_tbl[ bs-Dense1Pattern ];
-       if(bs<=Dense7Pattern)
-	   d=8;
-       else if (bs<=CrossPattern)
-	   d=24;
-       else
-	   d=16;
-   }
-   if ( bs == CustomPattern || pat ) {
+    uchar * pat=0;
+    int bs=cbrush.style();
+    int d=0;
+    if( bs>=Dense1Pattern && bs <= DiagCrossPattern ) {
+	pat=pat_tbl[ bs-Dense1Pattern ];
+	if(bs<=Dense7Pattern)
+	    d=8;
+	else if (bs<=CrossPattern)
+	    d=24;
+	else
+	    d=16;
+    }
+    if ( bs == CustomPattern || pat ) {
         QPixmap *pm;
         if ( pat ) {
             QString key="$qt-brush$" + QString::number( bs );
             pm = QPixmapCache::find( key );
             bool del = FALSE;
             if ( !pm ) {                        // not already in pm dict
-                pm = new QBitmap( d, d, pat, TRUE );
-                Q_CHECK_PTR( pm );
-                del = !QPixmapCache::insert( key, pm );
+		pm = new QBitmap( d, d, pat, TRUE );
+		Q_CHECK_PTR( pm );
+		del = !QPixmapCache::insert( key, pm );
             }
-            if ( cbrush.data->pixmap )
-                delete cbrush.data->pixmap;
-            cbrush.data->pixmap = new QPixmap( *pm );
-            if (del) delete pm;
-        }
-        pm = cbrush.data->pixmap;
-   }
+	    if ( cbrush.data->pixmap )
+		delete cbrush.data->pixmap;
+	    cbrush.data->pixmap = new QPixmap( *pm );
+	    if (del) delete pm;
+	}
+	pm = cbrush.data->pixmap;
+    }
 
-   gfx->setBrush( cbrush );
-   gfx->setBrushPixmap( cbrush.data->pixmap );
+    gfx->setBrush( cbrush );
+    gfx->setBrushPixmap( cbrush.data->pixmap );
 }
 
 
@@ -418,16 +422,16 @@ bool QPainter::begin( const QPaintDevice *pd, bool unclipped )
     flags = IsActive | DirtyFont;		// init flags
     int dt = pdev->devType();			// get the device type
 
-    if ( (pdev->devFlags & QInternal::ExternalDevice) != 0 )	// this is an extended device
+    if ( (pdev->devFlags & QInternal::ExternalDevice) != 0 ) // external device
 	setf(ExtDev);
     else if ( dt == QInternal::Pixmap )		// device is a pixmap
 	((QPixmap*)pdev)->detach();		// will modify it
 
-    gfx=((QPaintDevice *)pd)->graphicsContext();
-//  if gfx == 0 then we have an invalid pixmap
-    if ( gfx == 0 ) {
-#if defined(CHECK_NULL)
-	qWarning( "Unable to get graphics context" );
+    gfx = ((QPaintDevice *)pdev)->graphicsContext();
+//  if gfx == 0 we might have an invalid pixmap
+    if ( gfx == 0 && !testf(ExtDev) ) {
+#if defined(QT_CHECK_NULL)
+	qWarning( "QPainter::begin: Unable to get graphics context" );
 #endif
 	return FALSE;
     }
@@ -446,6 +450,7 @@ bool QPainter::begin( const QPaintDevice *pd, bool unclipped )
 
     pdev->painters++;				// also tell paint device
     bro = QPoint( 0, 0 );
+    d->curPt = QPoint( 0, 0 );
     if ( reinit ) {
 	bg_mode = TransparentMode;		// default background mode
 	rop = CopyROP;				// default ROP
@@ -484,14 +489,14 @@ bool QPainter::begin( const QPaintDevice *pd, bool unclipped )
 	ww = vw = w->width();			// default view size
 	wh = vh = w->height();
 #endif
-	if ( unclipped || w->testWFlags( WPaintUnclipped ) ) { // paint direct on device
+	if ( unclipped || w->testWFlags(WPaintUnclipped) ) { // paint direct on device
 	    setf( NoCache );
 	    updatePen();
 	    updateBrush();
-	    //gfx->setWidgetRegion( w->rect() );
-	    gfx->setWidgetRegion( QRect( 0, 0, qt_screen->width(), qt_screen->height() ) );
+	    if ( !testf(ExtDev) )
+		gfx->setWidgetRegion( QRect(w->mapToGlobal(QPoint(0,0)), w->size()) );
 	}
-
+	w->setActivePainter( this );
     } else if ( dt == QInternal::Pixmap ) {		// device is a pixmap
 	QPixmap *pm = (QPixmap*)pdev;
 	if ( pm->isNull() ) {
@@ -534,8 +539,12 @@ bool QPainter::begin( const QPaintDevice *pd, bool unclipped )
     updateBrush();
     updatePen();
 
-    if ( paintEventDevice == device() )
-	gfx->setClipRegion( *paintEventClipRegion );
+    if ( !testf(ExtDev) ) {
+	if ( paintEventDevice == device() )
+	    gfx->setClipRegion( *paintEventClipRegion );
+	else
+	    gfx->setClipping( FALSE );
+    }
 
     return TRUE;
 }
@@ -560,24 +569,37 @@ bool QPainter::end()				// end painting
     if ( paintEventSaveRegion )
 	*paintEventSaveRegion = QRegion();
 
+    if ( pdev->devType() == QInternal::Widget ) {
+	QWidget *w = (QWidget*)pdev;
+	w->clearActivePainter();
+    }
+
     delete gfx;
     gfx = 0;
 
+    if ( pfont ) {
+	delete pfont;
+	pfont = 0;
+    }
+
     flags = 0;
     pdev->painters--;
     pdev = 0;
     return TRUE;
 }
 
+
 void QPainter::flush(const QRegion &, CoordinateMode)
 {
     flush();
 }
 
+
 void QPainter::flush()
 {
 }
 
+
 void QPainter::setBackgroundColor( const QColor &c )
 {
     if ( !isActive() ) {
@@ -590,18 +612,13 @@ void QPainter::setBackgroundColor( const QColor &c )
     if ( testf(ExtDev) ) {
 	QPDevCmdParam param[1];
 	param[0].color = &bg_col;
-	if ( !pdev->cmd( QPaintDevice::PdcSetBkColor, this, param )  )
+	if ( !pdev->cmd(QPaintDevice::PdcSetBkColor,this,param) || !gfx )
 	    return;
     }
-    if ( !gfx )
-	return;
-    if ( !penRef )
-	updatePen();				// update pen setting
-    if ( !brushRef )
-	updateBrush();				// update brush setting
-    gfx->setBackgroundColor(bg_col);
+    gfx->setBackgroundColor( bg_col );
 }
 
+
 void QPainter::setBackgroundMode( BGMode m )
 {
     if ( !isActive() ) {
@@ -620,16 +637,10 @@ void QPainter::setBackgroundMode( BGMode m )
     if ( testf(ExtDev) ) {
 	QPDevCmdParam param[1];
 	param[0].ival = m;
-	pdev->cmd( QPaintDevice::PdcSetBkMode, this, param );
-	return;
+	if ( !pdev->cmd(QPaintDevice::PdcSetBkMode,this,param) || !gfx )
+	    return;
     }
-    if ( !gfx )
-	return;
-    if ( !penRef )
-	updatePen();				// update pen setting
-    if ( !brushRef )
-	updateBrush();				// update brush setting
-    gfx->setOpaqueBackground(bg_mode==OpaqueMode);
+    gfx->setOpaqueBackground( bg_mode==OpaqueMode );
 }
 
 void QPainter::setRasterOp( RasterOp r )
@@ -650,20 +661,13 @@ void QPainter::setRasterOp( RasterOp r )
     if ( testf(ExtDev) ) {
 	QPDevCmdParam param[1];
 	param[0].ival = r;
-	if ( !pdev->cmd( QPaintDevice::PdcSetROP, this, param ) )
+	if ( !pdev->cmd(QPaintDevice::PdcSetROP,this,param) || !gfx )
 	    return;
     }
-    if ( !gfx )
-	return;
-    if ( penRef )
-	updatePen();				// get non-cached pen GC
-    if ( brushRef )
-	updateBrush();				// get non-cached brush GC
-
-    gfx->setRop(r);
-
+    gfx->setRop( r );
 }
 
+
 void QPainter::setBrushOrigin( int x, int y )
 {
     if ( !isActive() ) {
@@ -676,13 +680,12 @@ void QPainter::setBrushOrigin( int x, int y )
     if ( testf(ExtDev) ) {
 	QPDevCmdParam param[1];
 	param[0].point = &bro;
-	pdev->cmd( QPaintDevice::PdcSetBrushOrigin, this, param );
-	return;
+	if ( !pdev->cmd(QPaintDevice::PdcSetBrushOrigin,this,param) || !gfx )
+	    return;
     }
-    if ( brushRef )
-	updateBrush();				// get non-cached brush GC
 }
 
+
 void QPainter::setClipping( bool enable )
 {
     if ( !isActive() ) {
@@ -705,8 +708,7 @@ void QPainter::setClipping( bool enable )
 	    else
 		*paintEventSaveRegion = crgn;
 	    crgn = *paintEventClipRegion;
-	}
-	else {
+	} else {
 	    if ( paintEventSaveRegion && !paintEventSaveRegion->isNull() )
 		crgn = *paintEventSaveRegion;
 	}
@@ -714,18 +716,16 @@ void QPainter::setClipping( bool enable )
 
     setf( ClipOn, enable );
 
-    if(testf(ExtDev)) {
+    if ( testf(ExtDev) ) {
 	QPDevCmdParam param[1];
 	param[0].ival = enable;
-	pdev->cmd( QPaintDevice::PdcSetClip, this, param );
-	return;
+	if ( !pdev->cmd(QPaintDevice::PdcSetClip, this, param) || !gfx )
+	    return;
     }
-    if ( !gfx )
-	return;
     if ( enable ) {
-	gfx->setClipRegion(crgn);
+	gfx->setClipRegion( crgn );
     } else {
-	gfx->setClipping(FALSE);
+	gfx->setClipping( FALSE );
     }
 }
 
@@ -758,17 +758,17 @@ void QPainter::setClipRegion( const QRegion &rgn, CoordinateMode m )
 	if ( paintEventSaveRegion )
 	    *paintEventSaveRegion = QRegion();
     }
-
-    QPDevCmdParam param[1];
-    param[0].rgn = &crgn;
-    if(testf(ExtDev)) {
-	if ( !pdev->cmd( QPaintDevice::PdcSetClipRegion, this, param ) )
-	    return; // device cannot clip
+    if ( testf(ExtDev) ) {
+	QPDevCmdParam param[1];
+	param[0].rgn = &crgn;
+	if ( !pdev->cmd(QPaintDevice::PdcSetClipRegion, this, param) || !gfx )
+	    return;
     }
     clearf( ClipOn );
     setClipping( TRUE );
 }
 
+
 void QPainter::drawPolyInternal( const QPointArray &a, bool close )
 {
     if ( a.size() < 2 || !gfx )
@@ -780,7 +780,7 @@ void QPainter::drawPolyInternal( const QPointArray &a, bool close )
     bool do_close = close && !(x1 == x2 && y1 == y2);
 
     if ( close && cbrush.style() != NoBrush ) {	// draw filled polygon
-	gfx->drawPolygon(a,false,0,a.size());
+	gfx->drawPolygon(a,FALSE,0,a.size());
 	if ( cpen.style() == NoPen ) {		// draw fake outline
 	    gfx->drawPolyline(a,0,a.size());
 	    if ( do_close )
@@ -794,6 +794,7 @@ void QPainter::drawPolyInternal( const QPointArray &a, bool close )
     }
 }
 
+
 void QPainter::drawPoint( int x, int y )
 {
     if ( !isActive() )
@@ -804,54 +805,51 @@ void QPainter::drawPoint( int x, int y )
 	    if ( testf(ExtDev) ) {
 		QPoint p( x, y );
 		param[0].point = &p;
-		pdev->cmd( QPaintDevice::PdcDrawPoint, this, param );
-		return;
+		if ( !pdev->cmd(QPaintDevice::PdcDrawPoint,this,param) || !gfx)
+		    return;
 	    }
 	    map( x, y, &x, &y );
 	}
-	if ( gfx )
-	    gfx->drawPoint(x,y);
+	gfx->drawPoint(x,y);
     }
 }
 
 
 void QPainter::drawPoints( const QPointArray& a, int index, int npoints )
 {
-
     if ( npoints < 0 )
 	npoints = a.size() - index;
     if ( index + npoints > (int)a.size() )
 	npoints = a.size() - index;
-    if ( !isActive() || npoints < 1 || index < 0 )
+    if ( !isActive() || npoints < 1 || index < 0 || cpen.style() == NoPen )
 	return;
     QPointArray pa = a;
-    if ( cpen.style() != NoPen ) {
-	if ( testf(ExtDev|VxF|WxF) ) {
-	    if ( testf(ExtDev) ) {
-		QPDevCmdParam param[1];
-		for (int i=0; i<npoints; i++) {
-		    QPoint p( pa[index+i].x(), pa[index+i].y() );
-		    param[0].point = &p;
-		    if ( !pdev->cmd( QPaintDevice::PdcDrawPoint, this, param ))
-			return;
-		}
-		return;
+    if ( testf(ExtDev|VxF|WxF) ) {
+	if ( testf(ExtDev) ) {
+	    QPDevCmdParam param[1];
+	    for (int i=0; i<npoints; i++) {
+		QPoint p( pa[index+i].x(), pa[index+i].y() );
+		param[0].point = &p;
+		if ( !pdev->cmd(QPaintDevice::PdcDrawPoint, this, param) )
+		    return;
 	    }
-	    bool tx = FALSE;
+	    if ( !gfx )
+		return;
+	}
 #ifndef QT_NO_TRANSFORMATIONS
-	    tx = (txop != TxNone);
+	bool tx = (txop != TxNone);
+#else
+	bool tx = xlatex || xlatey;
 #endif
-	    if ( tx ) {
-		pa = xForm( a, index, npoints );
-		if ( pa.size() != a.size() ) {
-		    index = 0;
-		    npoints = pa.size();
-		}
+	if ( tx ) {
+	    pa = xForm( a, index, npoints );
+	    if ( pa.size() != a.size() ) {
+		index = 0;
+		npoints = pa.size();
 	    }
 	}
-	if ( gfx )
-	    gfx->drawPoints(pa,index,npoints);
     }
+    gfx->drawPoints( pa, index, npoints );
 }
 
 
@@ -864,15 +862,16 @@ void QPainter::moveTo( int x, int y )
 	    QPDevCmdParam param[1];
 	    QPoint p( x, y );
 	    param[0].point = &p;
-	    pdev->cmd( QPaintDevice::PdcMoveTo, this, param );
-	    return;
+	    if ( !pdev->cmd(QPaintDevice::PdcMoveTo, this, param) || !gfx )
+		return;
 	}
 	map( x, y, &x, &y );
     }
-    if ( gfx )
-	gfx->moveTo(x,y);
+    gfx->moveTo( x, y );
+    d->curPt = QPoint( x, y );
 }
 
+
 void QPainter::lineTo( int x, int y )
 {
     if ( !isActive() )
@@ -882,15 +881,16 @@ void QPainter::lineTo( int x, int y )
 	    QPDevCmdParam param[1];
 	    QPoint p( x, y );
 	    param[0].point = &p;
-	    pdev->cmd( QPaintDevice::PdcLineTo, this, param );
-	    return;
+	    if ( !pdev->cmd(QPaintDevice::PdcLineTo, this, param) || !gfx )
+		return;
 	}
 	map( x, y, &x, &y );
     }
-    if ( gfx )
-	gfx->lineTo(x, y);
+    gfx->lineTo( x, y );
+    d->curPt = QPoint( x, y );
 }
 
+
 void QPainter::drawLine( int x1, int y1, int x2, int y2 )
 {
     if ( !isActive() )
@@ -901,23 +901,23 @@ void QPainter::drawLine( int x1, int y1, int x2, int y2 )
 	    QPoint p1(x1, y1), p2(x2, y2);
 	    param[0].point = &p1;
 	    param[1].point = &p2;
-	    pdev->cmd( QPaintDevice::PdcDrawLine, this, param );
-	    return;
+	    if ( !pdev->cmd(QPaintDevice::PdcDrawLine, this, param) || !gfx )
+		return;
 	}
 	map( x1, y1, &x1, &y1 );
 	map( x2, y2, &x2, &y2 );
     }
-    if ( cpen.style() != NoPen && gfx ) {
-	gfx->drawLine(x1,y1,x2,y2);
-    }
+    if ( cpen.style() != NoPen )
+	gfx->drawLine( x1, y1, x2, y2 );
+    gfx->moveTo( x2, y2 );
+    d->curPt = QPoint( x2, y2 );
 }
 
+
 void QPainter::drawRect( int x, int y, int w, int h )
 {
     if ( !isActive() )
 	return;
-
-#ifndef QT_NO_TRANSFORMATIONS
     if ( testf(ExtDev|VxF|WxF) ) {
 	if ( testf(ExtDev) ) {
 	    QPDevCmdParam param[1];
@@ -926,18 +926,15 @@ void QPainter::drawRect( int x, int y, int w, int h )
 	    if ( !pdev->cmd( QPaintDevice::PdcDrawRect, this, param ) || !gfx )
 		return;
 	}
+#ifndef QT_NO_TRANSFORMATIONS
 	if ( txop == TxRotShear ) {		// rotate/shear polygon
 	    QPointArray a( QRect(x,y,w,h), TRUE );
 	    drawPolyInternal( xForm(a) );
 	    return;
 	}
+#endif
 	map( x, y, w, h, &x, &y, &w, &h );
     }
-#else
-    map( x, y, w, h, &x, &y, &w, &h );
-#endif
-    if ( !gfx )
-	return;
     if ( w <= 0 || h <= 0 )
 	fix_neg_rect( &x, &y, &w, &h );
 
@@ -964,9 +961,10 @@ void QPainter::drawRect( int x, int y, int w, int h )
 	}
     }
 
-    gfx->fillRect(x,y,w,h);
+    gfx->fillRect( x, y, w, h );
 }
 
+
 void QPainter::drawWinFocusRect( int x, int y, int w, int h )
 {
     drawWinFocusRect( x, y, w, h, TRUE, color0 );
@@ -987,51 +985,42 @@ void QPainter::drawWinFocusRect( int x, int y, int w, int h,
     if ( txop == TxRotShear )
         return;
 #endif
-    static char winfocus_line[] = { 1, 1 };
 
-    QPen    old_pen = cpen;
-    QBrush  old_brush = cbrush;
+    if ( testf(ExtDev|VxF|WxF) ) {
+        if ( testf(ExtDev) ) {
+            QPDevCmdParam param[1];
+            QRect r( x, y, w, h );
+            param[0].rect = &r;
+            if ( !pdev->cmd(QPaintDevice::PdcDrawRect, this, param) || !gfx )
+		return;
+	}
+	map( x, y, w, h, &x, &y, &w, &h );
+    }
+
+    if ( w <= 0 || h <= 0 ) {
+        if ( w == 0 || h == 0 )
+            return;
+        fix_neg_rect( &x, &y, &w, &h );
+    }
+
+    static char winfocus_line[] = { 1, 1 };
+    QPen     old_pen = cpen;
+    QBrush   old_brush = cbrush;
     RasterOp old_rop = (RasterOp)rop;
 
     setBrush( QBrush() );
 
     if ( xorPaint ) {
         if ( QColor::numBitPlanes() <= 8 )
-            setPen( color1 );
-        else
-            setPen( white );
-        setRasterOp( XorROP );
+	    setPen( color1 );
+	else
+	    setPen( white );
+	setRasterOp( XorROP );
     } else {
-        if ( qGray( bgColor.rgb() ) < 128 )
-            setPen( white );
-        else
-            setPen( black );
-    }
-
-#ifndef QT_NO_TRANSFORMATIONS
-    if ( testf(ExtDev|VxF|WxF) ) {
-        if ( testf(ExtDev) ) {
-            QPDevCmdParam param[1];
-            QRect r( x, y, w, h );
-            param[0].rect = &r;
-            if ( !pdev->cmd( QPaintDevice::PdcDrawRect, this, param )) {
-                setRasterOp( old_rop );
-                setPen( old_pen );
-                return;
-            }
-        }
-        map( x, y, w, h, &x, &y, &w, &h );
-    }
-#else
-    map( x, y, w, h, &x, &y, &w, &h );
-#endif
-    if ( !gfx )
-	return;
-
-    if ( w <= 0 || h <= 0 ) {
-        if ( w == 0 || h == 0 )
-            return;
-        fix_neg_rect( &x, &y, &w, &h );
+	if ( qGray( bgColor.rgb() ) < 128 )
+	    setPen( white );
+	else
+	    setPen( black );
     }
 
     gfx->setDashes(winfocus_line, 2);
@@ -1054,6 +1043,7 @@ void QPainter::drawWinFocusRect( int x, int y, int w, int h,
     setBrush( old_brush );
 }
 
+
 void QPainter::drawRoundRect( int x, int y, int w, int h, int xRnd, int yRnd )
 {
 
@@ -1067,15 +1057,14 @@ void QPainter::drawRoundRect( int x, int y, int w, int h, int xRnd, int yRnd )
 	xRnd = 99;
     if ( yRnd >= 100 )
 	yRnd = 99;
-
     if ( testf(ExtDev) ) {
 	QPDevCmdParam param[3];
 	QRect r( x, y, w, h );
 	param[0].rect = &r;
 	param[1].ival = xRnd;
 	param[2].ival = yRnd;
-	pdev->cmd( QPaintDevice::PdcDrawRoundRect, this, param );
-	return;
+	if ( !pdev->cmd(QPaintDevice::PdcDrawRoundRect, this, param) || !gfx )
+	    return;
     }
     QPointArray a;
     if ( w <= 0 || h <= 0 )
@@ -1113,6 +1102,7 @@ void QPainter::drawRoundRect( int x, int y, int w, int h, int xRnd, int yRnd )
     drawPolyInternal( xForm(a) );
 }
 
+
 void QPainter::drawEllipse( int x, int y, int w, int h )
 {
     if ( !isActive() )
@@ -1121,8 +1111,8 @@ void QPainter::drawEllipse( int x, int y, int w, int h )
 	QPDevCmdParam param[1];
 	QRect r( x, y, w, h );
 	param[0].rect = &r;
-	pdev->cmd( QPaintDevice::PdcDrawEllipse, this, param );
-	return;
+	if ( !pdev->cmd(QPaintDevice::PdcDrawEllipse, this, param) || !gfx )
+	    return;
     }
     QPointArray a;
 #ifndef QT_NO_TRANSFORMATIONS
@@ -1150,15 +1140,15 @@ void QPainter::drawArc( int x, int y, int w, int h, int a, int alen )
 	param[0].rect = &r;
 	param[1].ival = a;
 	param[2].ival = alen;
-	pdev->cmd( QPaintDevice::PdcDrawArc, this, param );
-	return;
+	if ( !pdev->cmd(QPaintDevice::PdcDrawArc, this, param) || !gfx )
+	    return;
     }
     QPointArray pa;
 #ifndef QT_NO_TRANSFORMATIONS
-    pa.makeArc( x, y, w, h, a, alen, xmat ); // arc polyline
+    pa.makeArc( x, y, w, h, a, alen, xmat );	// arc polyline
 #else
     map( x, y, &x, &y );
-    pa.makeArc( x, y, w, h, a, alen ); // arc polyline
+    pa.makeArc( x, y, w, h, a, alen );		// arc polyline
 #endif
     drawPolyInternal( pa, FALSE );
 }
@@ -1173,7 +1163,6 @@ void QPainter::drawPie( int x, int y, int w, int h, int a, int alen )
 	a = a % (360*16);
 	if ( a < 0 ) a += (360*16);
     }
-
     if ( !isActive() )
 	return;
     if ( testf(ExtDev) ) {
@@ -1182,15 +1171,15 @@ void QPainter::drawPie( int x, int y, int w, int h, int a, int alen )
 	param[0].rect = &r;
 	param[1].ival = a;
 	param[2].ival = alen;
-	pdev->cmd( QPaintDevice::PdcDrawPie, this, param );
-	return;
+	if ( !pdev->cmd(QPaintDevice::PdcDrawPie, this, param) || !gfx )
+	    return;
     }
     QPointArray pa;
 #ifndef QT_NO_TRANSFORMATIONS
-    pa.makeArc( x, y, w, h, a, alen, xmat ); // arc polyline
+    pa.makeArc( x, y, w, h, a, alen, xmat );	// arc polyline
 #else
     map( x, y, &x, &y );
-    pa.makeArc( x, y, w, h, a, alen ); // arc polyline
+    pa.makeArc( x, y, w, h, a, alen );		// arc polyline
 #endif
     int n = pa.size();
     int cx, cy;
@@ -1201,7 +1190,7 @@ void QPainter::drawPie( int x, int y, int w, int h, int a, int alen )
     cy = y+h/2;
 #endif
     pa.resize( n+2 );
-    pa.setPoint( n, cx, cy );	// add legs
+    pa.setPoint( n, cx, cy );			// add legs
     pa.setPoint( n+1, pa.at(0) );
     drawPolyInternal( pa );
 }
@@ -1217,19 +1206,19 @@ void QPainter::drawChord( int x, int y, int w, int h, int a, int alen )
 	param[0].rect = &r;
 	param[1].ival = a;
 	param[2].ival = alen;
-	pdev->cmd(QPaintDevice::PdcDrawChord,this,param);
-	return;
+	if ( !pdev->cmd(QPaintDevice::PdcDrawChord,this,param) || !gfx )
+	    return;
     }
     QPointArray pa;
 #ifndef QT_NO_TRANSFORMATIONS
-    pa.makeArc( x, y, w-1, h-1, a, alen, xmat ); // arc polygon
+    pa.makeArc( x, y, w-1, h-1, a, alen, xmat );// arc polygon
 #else
     map( x, y, &x, &y );
-    pa.makeArc( x, y, w-1, h-1, a, alen ); // arc polygon
+    pa.makeArc( x, y, w-1, h-1, a, alen );	// arc polygon
 #endif
     int n = pa.size();
     pa.resize( n+1 );
-    pa.setPoint( n, pa.at(0) );		// connect endpoints
+    pa.setPoint( n, pa.at(0) );			// connect endpoints
     drawPolyInternal( pa );
 }
 
@@ -1252,24 +1241,32 @@ void QPainter::drawLineSegments( const QPointArray &a, int index, int nlines )
 	}
 	QPDevCmdParam param[1];
 	param[0].ptarr = (QPointArray*)&pa;
-	pdev->cmd(QPaintDevice::PdcDrawLineSegments,this,param);
-	return;
+	if ( !pdev->cmd(QPaintDevice::PdcDrawLineSegments,this,param) || !gfx )
+	    return;
     }
 #ifndef QT_NO_TRANSFORMATIONS
-    if ( txop != TxNone )
+    bool tx = (txop != TxNone);
+#else
+    bool tx = xlatex || xlatey;
 #endif
-	pa = xForm( pa, index, nlines*2 );
-    if ( pa.size() != a.size() ) {
-	index  = 0;
-	nlines = pa.size()/2;
-    }
-
+    int x1, y1, x2, y2;
     for ( int i=0; i<nlines; i++ ) {
-	drawLine(a[index],a[index+1]);
-	index+=2;
+	a.point( index++, &x1, &y1 );
+	a.point( index++, &x2, &y2 );
+        if ( tx ) {
+	    map( x1, y1, &x1, &y1 );
+	    map( x2, y2, &x2, &y2 );
+	}
+	if ( cpen.style() != NoPen )
+	    gfx->drawLine( x1, y1, x2, y2 );
+    }
+    if ( nlines > 0 ) {
+	gfx->moveTo( x2, y2 );
+	d->curPt = QPoint( x2, y2 );
     }
 }
 
+
 void QPainter::drawPolyline( const QPointArray &a, int index, int npoints )
 {
     if ( npoints < 0 )
@@ -1279,37 +1276,34 @@ void QPainter::drawPolyline( const QPointArray &a, int index, int npoints )
     if ( !isActive() || npoints < 2 || index < 0 )
 	return;
     QPointArray pa = a;
-    bool tx = TRUE;
+    if ( testf(ExtDev|VxF|WxF) ) {
+	if ( testf(ExtDev) ) {
+	    if ( npoints != (int)pa.size() ) {
+		pa = QPointArray( npoints );
+		for ( int i=0; i<npoints; i++ )
+		    pa.setPoint( i, a.point(index+i) );
+		index = 0;
+	    }
+	    QPDevCmdParam param[1];
+	    param[0].ptarr = (QPointArray*)&pa;
+	    if ( !pdev->cmd(QPaintDevice::PdcDrawPolyline,this,param) || !gfx )
+		return;
+	}
 #ifndef QT_NO_TRANSFORMATIONS
-    tx = (txop != TxNone);
+	bool tx = (txop != TxNone);
+#else
+	bool tx = xlatex || xlatey;
 #endif
-    if ( !testf(ExtDev) && tx ) {
-	pa = xForm( a, index, npoints );
-	if ( pa.size() != a.size() ) {
-	    index   = 0;
-	    npoints = pa.size();
+	if ( tx ) {
+	    pa = xForm( a, index, npoints );
+	    if ( pa.size() != a.size() ) {
+		index   = 0;
+		npoints = pa.size();
+	    }
 	}
     }
-    if ( npoints != (int)pa.size() ) {
-	pa = QPointArray( npoints );
-	for ( int i=0; i<npoints; i++ )
-	    pa.setPoint( i, a.point(index+i) );
-	index = 0;
-    }
-    if(testf(ExtDev)) {
-	QPDevCmdParam param[1];
-	param[0].ptarr = (QPointArray*)&pa;
-	pdev->cmd(QPaintDevice::PdcDrawPolyline,this,param);
-    } else if ( gfx ) {
-	gfx->drawPolyline(pa,index,npoints);
-    /*
-	int loopc;
-	moveTo(a[0]);
-	for(loopc=1;loopc<(int)a.count();loopc++) {
-	    lineTo(a[loopc]);
-	}
-    */
-    }
+    if ( cpen.style() != NoPen )
+	gfx->drawPolyline( pa, index, npoints );
 }
 
 
@@ -1320,6 +1314,7 @@ void QPainter::drawConvexPolygon( const QPointArray &pa,
     drawPolygon(pa,FALSE,index,npoints);
 }
 
+
 void QPainter::drawPolygon( const QPointArray &a, bool winding,
 			    int index, int npoints )
 {
@@ -1330,21 +1325,36 @@ void QPainter::drawPolygon( const QPointArray &a, bool winding,
     if ( !isActive() || npoints < 2 || index < 0 )
 	return;
     QPointArray pa = a;
-    bool tx = TRUE;
+    if ( testf(ExtDev|VxF|WxF) ) {
+	if ( testf(ExtDev) ) {
+	    if ( npoints != (int)a.size() ) {
+		pa = QPointArray( npoints );
+		for ( int i=0; i<npoints; i++ )
+		    pa.setPoint( i, a.point(index+i) );
+	    }
+	    QPDevCmdParam param[2];
+	    param[0].ptarr = (QPointArray*)&pa;
+	    param[1].ival = winding;
+	    if ( !pdev->cmd(QPaintDevice::PdcDrawPolygon,this,param) || !gfx )
+		return;
+	}
 #ifndef QT_NO_TRANSFORMATIONS
-    tx = (txop != TxNone);
+	bool tx = (txop != TxNone);
+#else
+	bool tx = xlatex || xlatey;
 #endif
-    if ( !testf(ExtDev) && tx ) {
-	pa = xForm( a, index, npoints );
-	if ( pa.size() != a.size() ) {
-	    index   = 0;
-	    npoints = pa.size();
+	if ( tx ) {
+	    pa = xForm( a, index, npoints );
+	    if ( pa.size() != a.size() ) {
+		index   = 0;
+		npoints = pa.size();
+	    }
 	}
     }
-    if ( gfx )
-	gfx->drawPolygon(pa,winding,index,npoints);
+    gfx->drawPolygon( pa, winding, index, npoints );
 }
 
+
 #ifndef QT_NO_BEZIER
 void QPainter::drawCubicBezier( const QPointArray &a, int index )
 {
@@ -1363,20 +1373,29 @@ void QPainter::drawCubicBezier( const QPointArray &a, int index )
 	for ( int i=0; i<4; i++ )
 	    pa.setPoint( i, a.point(index+i) );
     }
-#ifndef QT_NO_TRANSFORMATIONS
-    if ( !testf(ExtDev) ) {
-	if ( txop != TxNone ) {
-	    pa = xForm( pa );
-	    pa = pa.cubicBezier();
+    if ( testf(ExtDev|VxF|WxF) ) {
+	if ( testf(ExtDev) ) {
+	    QPDevCmdParam param[1];
+	    param[0].ptarr = (QPointArray*)&pa;
+	    if ( !pdev->cmd(QPaintDevice::PdcDrawCubicBezier,this,param)
+		 || !gfx )
+		return;
 	}
-    }
+#ifndef QT_NO_TRANSFORMATIONS
+	if ( txop != TxNone )
+#else
+	if ( xlatex || xlatey )
 #endif
-    QPDevCmdParam param[1];
-    param[0].ptarr = (QPointArray*)&pa;
-    pdev->cmd(QPaintDevice::PdcDrawCubicBezier,this,param);
+	    pa = xForm( pa );
+    }
+    if ( cpen.style() != NoPen ) {
+	pa = pa.cubicBezier();
+	gfx->drawPolyline( pa, 0, a.size() );
+    }
 }
 #endif //QT_NO_BEZIER
 
+
 void QPainter::drawPixmap( int x, int y, const QPixmap &pixmap,
 			   int sx, int sy, int sw, int sh )
 {
@@ -1429,9 +1448,10 @@ void QPainter::drawPixmap( int x, int y, const QPixmap &pixmap,
 		QRect r( x, y, pixmap.width(), pixmap.height() );
 		param[0].rect	= &r;
 		param[1].pixmap = &pixmap;
-		pdev->cmd(QPaintDevice::PdcDrawPixmap,this,param);
-		return;
+		if ( !pdev->cmd(QPaintDevice::PdcDrawPixmap,this,param) || !gfx )
+		    return;
 	    }
+#ifndef QT_NO_PIXMAP_TRANSFORMATION
 	    if ( txop == TxScale || txop == TxRotShear ) {
 		QWMatrix mat( m11(), m12(),
 			      m21(), m22(),
@@ -1452,17 +1472,26 @@ void QPainter::drawPixmap( int x, int y, const QPixmap &pixmap,
 		flags = save_flags;
 		return;
 	    }
+#endif
 	}
 	map( x, y, &x, &y );
     }
 #else
-    map( x, y, &x, &y );
+    if ( testf(ExtDev|VxF|WxF) ) {
+	if ( testf(ExtDev) ) {
+	    QPDevCmdParam param[2];
+	    QRect r( x, y, pixmap.width(), pixmap.height() );
+	    param[0].rect   = &r;
+	    param[1].pixmap = &pixmap;
+	    if ( !pdev->cmd(QPaintDevice::PdcDrawPixmap,this,param) || !gfx )
+		return;
+	}
+	map( x, y, &x, &y );
+    }
 #endif
-    if ( !gfx )
-	return;
     //bitBlt( pdev, x, y, &pixmap, sx, sy, sw, sh, CopyROP );
     gfx->setSource(&pixmap);
-    if(sw>pixmap.width()) {
+    if(sw>pixmap.width()) { // ###hanord: isn't this done above?
 	sw=pixmap.width();
     }
     if(sh>pixmap.height()) {
@@ -1482,6 +1511,7 @@ void QPainter::drawPixmap( int x, int y, const QPixmap &pixmap,
     gfx->blt(x,y,sw,sh,sx,sy);
 }
 
+
 void QPainter::drawTiledPixmap( int x, int y, int w, int h,
 				const QPixmap &pixmap, int sx, int sy )
 {
@@ -1516,6 +1546,7 @@ void QPainter::drawTiledPixmap( int x, int y, int w, int h,
     gfx->tiledBlt(x,y,w,h);
 }
 
+
 #ifndef QT_NO_TRANSFORMATIONS
 
 //
@@ -1562,16 +1593,18 @@ static void ins_text_bitmap( const QString &key, QBitmap *bm )
 
 #endif // QT_NO_TRANSFORMATIONS
 
-void QPainter::drawText( int x, int y, const QString &str, int len, QPainter::TextDirection dir )
+
+void QPainter::drawText( int x, int y, const QString &str, int len,
+			 QPainter::TextDirection dir )
 {
     drawText( x, y, str, 0, len, dir );
 }
 
-void QPainter::drawText( int x, int y, const QString &str, int from, int len, QPainter::TextDirection dir)
+void QPainter::drawText( int x, int y, const QString &str, int from, int len,
+			 QPainter::TextDirection dir)
 {
-    if(memorymanager->fontAscent(cfont.handle())==0) {
+    if ( memorymanager->fontAscent(cfont.handle()) == 0 )
 	return;
-    }
 
     if ( !isActive() )
 	return;
@@ -1593,12 +1626,24 @@ void QPainter::drawText( int x, int y, const QString &str, int from, int len, QP
 	    updateFont();
 
 	if ( testf(ExtDev) ) {
-	    QPDevCmdParam param[2];
-	    QPoint p( x, y );
-	    param[0].point = &p;
-	    param[1].str = &shaped;
-	    pdev->cmd(QPaintDevice::PdcDrawText2,this,param);
-	    return;
+	    QPDevCmdParam param[3];
+	    QFontPrivate::TextRun *cache = new QFontPrivate::TextRun();
+	    pfont->d->textWidth( shaped, 0, len, cache ); // create cache
+	    bool retval = FALSE;
+	    QFontPrivate::TextRun *runs = cache;
+	    while ( cache ) {
+		QPoint p( x + cache->xoff, y + cache->yoff );
+		QString s =
+		    QConstString( cache->string, cache->length ).string();
+		param[0].point = &p;
+		param[1].str = &s;
+		param[2].ival = cache->script;
+		retval = pdev->cmd(QPaintDevice::PdcDrawText2, this, param);
+		cache = cache->next;
+	    }
+	    delete runs;
+	    if ( !retval || !gfx )
+		return;
 	}
 #ifndef QT_NO_TRANSFORMATIONS
 	if ( txop >= TxScale ) {
@@ -1626,10 +1671,12 @@ void QPainter::drawText( int x, int y, const QString &str, int from, int len, QP
 		double rx = mat1.m11() * cfont.pointSizeFloat() / newSize;
 		double ry = mat1.m22() * cfont.pointSizeFloat() / newSize;
 		mat2 = QWMatrix( rx, 0, 0, ry, 0, 0 );
+#ifndef QT_NO_PIXMAP_TRANSFORMATION
 	    } else {
 		mat2 = QPixmap::trueMatrix( mat1, w, h );
 		aw = w;
 		ah = h;
+#endif
 	    }
 	    bool empty = aw == 0 || ah == 0;
 	    QPixmap *tpm = 0;
@@ -1657,7 +1704,11 @@ void QPainter::drawText( int x, int y, const QString &str, int from, int len, QP
 			p++;
 		    }
 		}
+#ifndef QT_NO_PIXMAP_TRANSFORMATION
 		tpm = new QPixmap( pm.xForm( mat2 ) );
+#else
+		tpm = new QPixmap( pm );
+#endif
 		if ( tpm->isNull() ) {
 		    delete tpm;
 		    return;
@@ -1666,7 +1717,7 @@ void QPainter::drawText( int x, int y, const QString &str, int from, int len, QP
 		bm_key = gen_text_bitmap_key( mat2, dfont, shaped, len );
 		wx_bm = get_text_bitmap( bm_key );
 		create_new_bm = wx_bm == 0;
-		if ( create_new_bm && !empty ) {	// no such cached bitmap
+		if ( create_new_bm && !empty ) {// no such cached bitmap
 		    QBitmap bm( aw, ah, TRUE );	// create bitmap
 		    QPainter paint;
 		    paint.begin( &bm );		// draw text in bitmap
@@ -1674,7 +1725,11 @@ void QPainter::drawText( int x, int y, const QString &str, int from, int len, QP
 		    paint.setFont( dfont );
 		    paint.drawText( tx, ty, shaped, len );
 		    paint.end();
+#ifndef QT_NO_PIXMAP_TRANSFORMATION
 		    wx_bm = new QBitmap( bm.xForm(mat2) ); // transform bitmap
+#else
+		    wx_bm = new QBitmap( bm );
+#endif
 		    if ( wx_bm->isNull() ) {
 			delete wx_bm;		// nothing to draw
 			return;
@@ -1696,12 +1751,9 @@ void QPainter::drawText( int x, int y, const QString &str, int from, int len, QP
 		QBrush oldBrush = cbrush;
 		setBrush( backgroundColor() );
 		updateBrush();
-		/* XXX
-		XFillPolygon( dpy, hd, gc_brush, (XPoint*)a.shortPoints(), 4,
-			      Nonconvex, CoordModeOrigin );
-		XDrawLines( dpy, hd, gc_brush, (XPoint*)a.shortPoints(), 5,
-			    CoordModeOrigin );
-		*/
+
+		drawPolygon( a, TRUE, 0, 4 );
+
 		setBrush( oldBrush );
 	    }
 	    if ( empty )
@@ -1713,9 +1765,6 @@ void QPainter::drawText( int x, int y, const QString &str, int from, int len, QP
 	    x = qRound(nfx-dx);
 	    y = qRound(nfy-dy);
 
-	    if ( !gfx )
-		return;
-
 	    if ( memorymanager->fontSmooth(dfont.handle()) &&
 		 QPaintDevice::qwsDisplay()->supportsDepth(32) ) {
 		gfx->setSource( tpm );
@@ -1736,28 +1785,31 @@ void QPainter::drawText( int x, int y, const QString &str, int from, int len, QP
 	}
 	if ( txop == TxTranslate )
 	    map( x, y, &x, &y );
+
+#else
+
+	map( x, y, &x, &y );
+
 #endif
     }
 
-#if defined(QT_NO_TRANSFORMATIONS)
-    map( x, y, &x, &y );
-#endif
+    QRect bbox;
+    int   lw, underlinepos, strikeoutpos, ascent;
+    QString newstr = str.left(len);
 
-    /* XXX
-    if ( bg_mode == TransparentMode )
-	XDrawString16( dpy, hd, gc, x, y, (XChar2b*)v.unicode(), len );
-    else
-	XDrawImageString16( dpy, hd, gc, x, y, (XChar2b*)v.unicode(), len );
-    */
-    /*
-    QPDevCmdParam param[2];
-    QPoint p( x, y );
-    param[0].point = &p;
-    param[1].str = &newstr;
-    pdev->cmd(QPaintDevice::PdcDrawText2,this,param);
-    */
-    if ( !gfx )
-	return;
+    if( cfont.underline() || cfont.strikeOut() || bg_mode == OpaqueMode ) {
+	QFontMetrics fm = fontMetrics();
+	lw = fm.lineWidth();
+	underlinepos = fm.underlinePos();
+	strikeoutpos = fm.strikeOutPos();
+	ascent = fm.ascent();
+	bbox = fm.boundingRect( newstr, len );
+	if ( bg_mode == OpaqueMode ) {		// opaque: fill background
+	    gfx->setBrush( QBrush(backgroundColor()) );
+	    gfx->fillRect( x, y-ascent, bbox.width(), bbox.height() );
+	    gfx->setBrush( cbrush );
+	}
+    }
 
     QFontPrivate::TextRun *cache = new QFontPrivate::TextRun();
     int width=cfont.d->textWidth( shaped, 0, len, cache );
@@ -1765,28 +1817,19 @@ void QPainter::drawText( int x, int y, const QString &str, int from, int len, QP
     delete cache;
 
     if ( cfont.underline() || cfont.strikeOut() ) {
-        QFontMetrics fm = fontMetrics();
-        int lw = fm.lineWidth();
-
-	gfx->save();
-	gfx->setBrush(cpen.color());
-
-        // draw underline effect
-        if ( cfont.underline() ) {
-	    gfx->fillRect(x,y+fm.underlinePos(), width, lw);
-        }
-
-        // draw strikeout effect
-        if ( cfont.strikeOut() ) {
-	    gfx->fillRect(x,y-fm.strikeOutPos(), width, lw);
-	}
-
-	gfx->restore();
-
+	gfx->setBrush( cpen.color() );
+        if ( cfont.underline() )		// draw underline effect
+	    gfx->fillRect( x, y+underlinepos, width, lw );
+        if ( cfont.strikeOut() )		// draw strikeout effect
+	    gfx->fillRect( x, y-strikeoutpos, width, lw );
+	gfx->setBrush( cbrush );
     }
 }
 
+
 QPoint QPainter::pos() const
 {
+    if ( !gfx )
+	return xFormDev( d->curPt );
     return gfx->pos();
 }
