@@ -1,5 +1,5 @@
 /****************************************************************************
-** $Id:  qt/qpixmap_qws.cpp   3.0.3   edited Feb 28 05:53 $
+** $Id: qt/qpixmap_qws.cpp   3.1.2   edited Jan 21 04:53 $
 **
 ** Implementation of QPixmap class for FB
 **
@@ -38,9 +38,11 @@
 #include "qwmatrix.h"
 #include "qapplication.h"
 #include "qpainter.h"
+#include "qptrdict.h"
 #include "qwsdisplay_qws.h"
 #include "qgfx_qws.h"
 #include <stdlib.h>
+#include <limits.h>
 
 #include "qmemorymanager_qws.h"
 
@@ -141,6 +143,96 @@ static void build_scale_table( uint **table, uint nBits )
 }
 */
 
+static QList<QShared> *qws_pixmapData = 0;
+static bool qws_trackPixmapData = TRUE;
+
+class QwsPixmap : public QPixmap
+{
+public:
+    QwsPixmap() : QPixmap() {}
+    static void mapPixmaps( bool from );
+    static void freeSharedData();
+    static QPtrDict<QImage> *images;
+};
+
+QPtrDict<QImage> *QwsPixmap::images = 0;
+
+void QwsPixmap::mapPixmaps( bool from )
+{
+    if ( !qws_pixmapData )
+	return;
+    if ( !images )
+	images = new QPtrDict<QImage>;
+    qws_trackPixmapData = FALSE;
+    QListIterator<QShared> it( *qws_pixmapData );
+    while ( it.current() ) {
+	QPixmapData *d = (QPixmapData*)it.current();
+	++it;
+	if ( d->w && d->h ) {
+	    if ( from ) {
+		QwsPixmap p;
+		QPixmapData *tmp = p.data;
+		p.data = d;
+		QImage *img = new QImage(p.convertToImage());
+		images->insert( d, img );
+		p.data = tmp;
+	    } else {
+		QImage *img = images->take( d );
+		if ( img ) {
+		    if ( d->clut )
+			delete [] d->clut;
+		    if ( memorymanager )
+			memorymanager->deletePixmap(d->id);
+		    QwsPixmap p;
+		    p.convertFromImage( *img );
+		    int cnt = d->count-1;
+		    p.data->mask = d->mask;
+		    *d = *p.data;
+		    while (cnt > 0) {
+			d->ref();
+			--cnt;
+		    }
+		    delete img;
+		    delete p.data;
+		    p.data = 0;
+		}
+	    }
+	}
+    }
+    if ( !from )
+	images->clear();
+    qws_trackPixmapData = TRUE;
+}
+
+// On exit it is possible that pixmap data is stored in the vram cache.
+// We need to free this memory to ensure the cache doesn't fill up
+// with unused data.
+void QwsPixmap::freeSharedData()
+{
+    if ( !qws_pixmapData || !memorymanager )
+	return;
+    QListIterator<QShared> it( *qws_pixmapData );
+    while ( it.current() ) {
+	QPixmapData *d = (QPixmapData*)it.current();
+	++it;
+	if ( d->w && d->h && memorymanager->inVRAM(d->id) ) {
+	    memorymanager->deletePixmap(d->id);
+	    d->w = d->h = 0;
+	    d->id = 0;
+	}
+    }
+}
+
+void qws_mapPixmaps( bool from )
+{
+    QwsPixmap::mapPixmaps( from );
+}
+
+void qws_freePixmapData()
+{
+    QwsPixmap::freeSharedData();
+}
+
 /*****************************************************************************
   QPixmap member functions
  *****************************************************************************/
@@ -150,12 +242,18 @@ void QPixmap::init( int w, int h, int d, bool bitmap, Optimization optim )
     static int serial = 0;
     int dd = defaultDepth();
 
+    if ( !qws_pixmapData )
+	qws_pixmapData = new QList<QShared>;
+
     if ( optim == DefaultOptim )		// use default optimization
 	optim = defOptim;
 
     data = new QPixmapData;
     Q_CHECK_PTR( data );
 
+    if ( qws_trackPixmapData )
+	qws_pixmapData->append( data );
+
     memset( data, 0, sizeof(QPixmapData) );
     data->id=0;
     data->count  = 1;
@@ -213,6 +311,8 @@ void QPixmap::init( int w, int h, int d, bool bitmap, Optimization optim )
 void QPixmap::deref()
 {
     if ( data && data->deref() ) {			// last reference lost
+	if ( qws_trackPixmapData )
+	    qws_pixmapData->removeRef( data );
 	if ( data->mask )
 	    delete data->mask;
 	if ( data->clut )
@@ -357,6 +457,7 @@ QImage QPixmap::convertToImage() const
     int	d  = depth();
     bool mono = d == 1;
 
+    const QBitmap* msk = mask();
 
     if( d == 15 || d == 16 ) {
 #ifndef QT_NO_QWS_DEPTH_16
@@ -367,9 +468,26 @@ QImage QPixmap::convertToImage() const
 	    register uint *p = (uint *)image.scanLine(y);
 	    ushort  *s = (ushort*)scanLine(y);
 	    uint *end = p + w;
-	    while ( p < end )
-		*p++ = qt_conv16ToRgb( *s++ );
+	    if ( msk ) {
+		uchar* a = msk->scanLine(y);
+		uchar bit = 1; // mask is LittleEndian
+		while ( p < end ) {
+		    uint rgb = qt_conv16ToRgb( *s++ );
+		    if ( !(*a & bit) )
+			rgb &= 0x00ffffff;
+		    *p++ = rgb;
+		    if (!(bit <<= 1)) {
+			++a;
+			bit = 1;
+		    }
+		}
+	    } else {
+		while ( p < end )
+		    *p++ = qt_conv16ToRgb( *s++ );
+	    }
 	}
+	if ( msk )
+	    image.setAlphaBuffer( TRUE );
 #endif
     } else {
 	// We can only create little-endian pixmaps
@@ -401,6 +519,75 @@ QImage QPixmap::convertToImage() const
 	image.setNumColors( numCols() );
 	for ( int i = 0; i < numCols(); i++ )
 	    image.setColor( i, clut()[i] );
+	if (mask()) {				// which pixels are used?
+	    QImage alpha = mask()->convertToImage();
+	    alpha = qt_screen->mapToDevice( alpha );
+	    bool ale = alpha.bitOrder() == QImage::LittleEndian;
+	    register uchar *p;
+	    int  used[256];
+	    memset( used, 0, sizeof(int)*256 );
+	    for ( int i=0; i<h; i++ ) {
+		uchar* asrc = alpha.scanLine( i );
+		p = image.scanLine( i );
+		for ( int x = 0; x < w; x++ ) {
+		    if ( ale ) {
+			if (asrc[x >> 3] & (1 << (x & 7)))
+			    used[*p]++;
+		    } else {
+			if (asrc[x >> 3] & (1 << (7 -(x & 7))))
+			    used[*p]++;
+		    }
+		    ++p;
+		}
+	    }
+
+	    int trans=0;
+	    int bestn=INT_MAX;
+	    for ( int i=0; i<numCols(); i++ ) {
+		if ( used[i] < bestn ) {
+		    bestn = used[i];
+		    trans = i;
+		    if ( !bestn )
+			break;
+		}
+	    }
+
+	    image.setColor( trans, image.color(trans)&0x00ffffff );
+
+	    for ( int i=0; i<h; i++ ) {
+		uchar* asrc = alpha.scanLine( i );
+		p = image.scanLine( i );
+		for ( int x = 0; x < w; x++ ) {
+		    if ( ale ) {
+			if (!(asrc[x >> 3] & (1 << (x & 7))))
+			    *p = trans;
+		    } else {
+			if (!(asrc[x >> 3] & (1 << (7 -(x & 7)))))
+			    *p = trans;
+		    }
+		    ++p;
+		}
+	    }
+	    image.setAlphaBuffer( TRUE );
+	}
+    } else if ( d == 32 && mask() ) {
+	QImage alpha = mask()->convertToImage();
+	bool ale = alpha.bitOrder() == QImage::LittleEndian;
+	for ( int i=0; i<h; i++ ) {
+	    uchar* asrc = alpha.scanLine( i );
+	    Q_UINT32 *p = (Q_UINT32 *)image.scanLine( i );
+	    for ( int x = 0; x < w; x++ ) {
+		if ( ale ) {
+		    if (!(asrc[x >> 3] & (1 << (x & 7))))
+			*p = *p & 0x00ffffff;
+		} else {
+		    if (!(asrc[x >> 3] & (1 << (7 -(x & 7)))))
+			*p = *p & 0x00ffffff;
+		}
+		++p;
+	    }
+	}
+	image.setAlphaBuffer( TRUE );
     }
 
     image = qt_screen->mapFromDevice( image );
@@ -446,7 +633,9 @@ bool QPixmap::convertFromImage( const QImage &img, int conversion_flags )
 	    if ( image.numColors() == 2 ) {
 		QRgb c0 = image.color(0);	// Auto: convert to best
 		QRgb c1 = image.color(1);
-		conv8 = QMIN(c0,c1) != 0 || QMAX(c0,c1) != qRgb(255,255,255);
+		conv8 = QMIN(c0,c1) != qRgb(0,0,0) || QMAX(c0,c1) != qRgb(255,255,255);
+		if ( !conv8 )
+		    force_mono = TRUE;
 	    } else {
 		// eg. 1-color monochrome images (they do exist).
 		conv8 = TRUE;
@@ -499,13 +688,16 @@ bool QPixmap::convertFromImage( const QImage &img, int conversion_flags )
 	    dd=32;
     }
 
+    // for drivers that do not accelerate alpha blt
+    Optimization optim = partialalpha ? NoOptim : defOptim;
+
     QImage rimg = qt_screen->mapToDevice( image );
 
     // detach other references and re-init()
     bool ibm = isQBitmap();
     detach();
     deref();
-    init( w, h, dd, ibm, defOptim );
+    init( w, h, dd, ibm, optim );
 
     QGfx * mygfx=graphicsContext();
     if ( mygfx ) {
@@ -574,44 +766,47 @@ QPixmap QPixmap::xForm( const QWMatrix &matrix ) const
     ws = width();
     hs = height();
 
-    QWMatrix mat = trueMatrix( matrix, ws, hs ); // true matrix
+    QWMatrix mat( matrix.m11(), matrix.m12(), matrix.m21(), matrix.m22(), 0., 0. );
 
-    // calculate new width and height
-    QPointArray a( QRect(0,0,ws,hs) );
-    a = mat.map( a );
-    QRect r = a.boundingRect().normalize();
-    w = r.width();
-    h = r.height();
-
-    if ( matrix.m12() == 0.0F  && matrix.m21() == 0.0F &&
-	 matrix.m11() >= 0.0F  && matrix.m22() >= 0.0F &&
-	 depth() == defaultDepth() // ### stretchBlt limitation
-	)
-    {
-	if ( mat.m11() == 1.0F && mat.m22() == 1.0F )
+    if ( matrix.m12() == 0.0F && matrix.m21() == 0.0F ) {
+	if ( matrix.m11() == 1.0F && matrix.m22() == 1.0F )
 	    return *this;			// identity matrix
+	h = qRound( matrix.m22()*hs );
+	w = qRound( matrix.m11()*ws );
+	h = QABS( h );
+	w = QABS( w );
+	if ( matrix.m11() >= 0.0F  && matrix.m22() >= 0.0F &&
+		depth() == defaultDepth() ) // ### stretchBlt limitation
+	{
+	     if ( w==0 || h==0 )
+		 return *this;
 
-	if(w==0 || h==0) {
-	    return *this;
-	}
-
-	QPixmap pm( w, h, depth(), NormalOptim );
-	QGfx * mygfx=pm.graphicsContext();
-	if ( mygfx ) {
-	    mygfx->setSource(this);
-	    mygfx->setAlphaType(QGfx::IgnoreAlpha);
-	    mygfx->stretchBlt(0,0,w,h,ws,hs);
-	    delete mygfx;
-	}
-	if ( data->mask ) {
-	    QBitmap bm =
-		data->selfmask ? *((QBitmap*)(&pm)) :
-					 data->mask->xForm(matrix);
-	    pm.setMask( bm );
-	}
-	return pm;
-
+	     QPixmap pm( w, h, depth(), NormalOptim );
+	     QGfx * mygfx=pm.graphicsContext();
+	     if ( mygfx ) {
+		 mygfx->setSource(this);
+		 mygfx->setAlphaType(QGfx::IgnoreAlpha);
+		 mygfx->stretchBlt(0,0,w,h,ws,hs);
+		 delete mygfx;
+	     }
+	     if ( data->mask ) {
+		 QBitmap bm =
+		     data->selfmask ? *((QBitmap*)(&pm)) :
+		     data->mask->xForm(matrix);
+		 pm.setMask( bm );
+	     }
+	     return pm;
+	 }
+    } else {					// rotation or shearing
+	QPointArray a( QRect(0,0,ws+1,hs+1) );
+	a = mat.map( a );
+	QRect r = a.boundingRect().normalize();
+	w = r.width()-1;
+	h = r.height()-1;
     }
+
+    mat = trueMatrix( mat, ws, hs ); // true matrix
+
     bool invertible;
     mat = mat.invert( &invertible );		// invert matrix
 
@@ -760,3 +955,9 @@ int QPixmap::numCols() const
 {
     return data->numcols;
 }
+
+bool QPixmap::hasAlpha() const
+{
+    return data->hasAlpha || data->mask;
+}
+
