@@ -1,5 +1,5 @@
 /****************************************************************************
-** $Id:  qt/qwidget_qws.cpp   3.0.3   edited Feb 20 19:20 $
+** $Id: qt/qwidget_qws.cpp   3.1.2   edited Jan 23 23:39 $
 **
 ** Implementation of QWidget and QWindow classes for FB
 **
@@ -46,11 +46,14 @@
 #include "qlayout.h"
 #include "qtextcodec.h"
 #include "qcursor.h"
+#include "qptrdict.h"
+#include "qinputcontext_p.h"
 
 #include "qwsdisplay_qws.h"
 #include "qgfx_qws.h"
 #include "qwsmanager_qws.h"
 #include "qwsregionmanager_qws.h"
+#include "qinputcontext_p.h"
 
 void qt_insert_sip( QWidget*, int, int );	// defined in qapplication_x11.cpp
 int  qt_sip_count( QWidget* );			// --- "" ---
@@ -64,6 +67,9 @@ extern bool qt_xdnd_enable( QWidget* w, bool on );
 
 extern void qt_deferred_map_add( QWidget* ); // defined in qapplication_x11.const
 extern void qt_deferred_map_take( QWidget* );// defined in qapplication_x11.const
+extern int qt_last_x;
+extern int qt_last_y;
+extern WId qt_last_cursor;
 
 extern QRect qt_maxWindowRect;
 
@@ -105,6 +111,30 @@ static void paint_children(QWidget * p,const QRegion& r, bool post)
     }
 }
 
+// Paint the widget and its children
+
+static void paint_heirarchy(QWidget *w, bool post)
+{
+    if ( w && w->testWState(Qt::WState_Visible) ) {
+	if ( post )
+	    QApplication::postEvent(w,new QPaintEvent(w->rect(),
+			!w->testWFlags(QWidget::WRepaintNoErase) ) );
+	else
+	    w->repaint(w->rect(),
+		    !w->testWFlags(QWidget::WRepaintNoErase));
+
+	QObjectList *childObjects=(QObjectList*)w->children();
+	if ( childObjects ) {
+	    QObject * o;
+	    for(o=childObjects->first();o!=0;o=childObjects->next()) {
+		if( o->isWidgetType() ) {
+		    paint_heirarchy((QWidget *)o,post);
+		}
+	    }
+	}
+    }
+}
+
 /*****************************************************************************
   QWidget member functions
  *****************************************************************************/
@@ -117,16 +147,12 @@ void QWidget::create( WId window, bool initializeWindow, bool /*destroyOldWindow
 
     if ( !parentWidget() || parentWidget()->isDesktop() )
 	setWFlags( WType_TopLevel );		// top-level widget
-    else if ( parentWidget()->isVisible() )
-	parentWidget()->setChildrenAllocatedDirty();
 
     alloc_region_index = -1;
     alloc_region_revision = -1;
     isSettingGeometry = FALSE;
     overlapping_children = -1;
 
-    static int sw = -1, sh = -1;		// screen size
-
     bool topLevel = testWFlags(WType_TopLevel);
     bool popup = testWFlags(WType_Popup);
     bool dialog = testWFlags(WType_Dialog);
@@ -149,10 +175,8 @@ void QWidget::create( WId window, bool initializeWindow, bool /*destroyOldWindow
 	    setWFlags(WStyle_StaysOnTop);
     }
 
-    if ( sw < 0 ) {				// get the screen size
-	sw = dpy->width();
-	sh = dpy->height();
-    }
+    int sw = dpy->width();
+    int sh = dpy->height();
 
     if ( dialog || popup || desktop ) {		// these are top-level, too
 	topLevel = TRUE;
@@ -207,6 +231,9 @@ void QWidget::create( WId window, bool initializeWindow, bool /*destroyOldWindow
 	}
     }
 
+    alloc_region_dirty=FALSE;
+    paintable_region_dirty=FALSE;
+
     if ( !initializeWindow ) {
 	// do no initialization
     } else if ( popup ) {			// popup widget
@@ -250,6 +277,7 @@ void QWidget::create( WId window, bool initializeWindow, bool /*destroyOldWindow
 	    qwsDisplay()->nameRegion( winId(), name(), caption() );
 	else
 	    qwsDisplay()->nameRegion( winId(), "", caption() );
+	createTLExtra();
 #else
 	qwsDisplay()->nameRegion( winId(), name(), QString::null );
 #endif
@@ -260,7 +288,11 @@ void QWidget::create( WId window, bool initializeWindow, bool /*destroyOldWindow
 	    // get size of wm decoration
 	    QRegion r = QApplication::qwsDecoration().region(this, crect);
 	    QRect br( r.boundingRect() );
-	    crect.moveBy( crect.x() - br.x(), crect.y() - br.y() );
+	    crect.moveBy( crect.x()-br.x(), crect.y()-br.y() );
+	    extra->topextra->fleft = crect.x()-br.x();
+	    extra->topextra->ftop = crect.y()-br.y();
+	    extra->topextra->fright = br.right()-crect.right();
+	    extra->topextra->fbottom = br.bottom()-crect.bottom();
 	    topData()->qwsManager = new QWSManager(this);
 	}
 #endif
@@ -272,14 +304,21 @@ void QWidget::create( WId window, bool initializeWindow, bool /*destroyOldWindow
 	// If we are session managed, inform the window manager about it
 	if ( extra && !extra->mask.isNull() ) {
 	    req_region = extra->mask;
-	    req_region.translate(crect.x(),crect.y()); //###expensive?
+	    req_region.translate(crect.x(),crect.y());
 	    req_region &= crect; //??? this is optional
 	} else {
 	    req_region = crect;
 	}
 	req_region = qt_screen->mapToDevice( req_region, QSize(qt_screen->width(), qt_screen->height()) );
+    } else {
+	if ( extra && extra->topextra )	{ // already allocated due to reparent?
+	    extra->topextra->fleft = 0;
+	    extra->topextra->ftop = 0;
+	    extra->topextra->fright = 0;
+	    extra->topextra->fbottom = 0;
+	}
+	updateRequestedRegion( mapToGlobal(QPoint(0,0)) );
     }
-    //### frame rect is not handled in Qt/Embedded
 }
 
 
@@ -340,17 +379,18 @@ void QWidget::reparentSys( QWidget *parent, WFlags f, const QPoint &p,
 
     setWinId( 0 );
 
-    QWidget *oldparent = parentWidget();
-    if ( oldparent ) {				// remove from parent
-	oldparent->removeChild( this );
-	oldparent->setChildrenAllocatedDirty();
-	oldparent->paintable_region_dirty = TRUE;
-    }
-    if ( parent ) {				// insert into new parent
-	parentObj = parent;			// avoid insertChild warning
-	parent->insertChild( this );
-	parent->setChildrenAllocatedDirty();
-	parent->paintable_region_dirty = TRUE;
+    if ( parentObj != parent ) {
+	QWidget *oldparent = parentWidget();
+	if ( oldparent ) {				// remove from parent
+	    oldparent->removeChild( this );
+	    oldparent->setChildrenAllocatedDirty();
+	    oldparent->paintable_region_dirty = TRUE;
+	}
+	if ( parent ) {				// insert into new parent
+	    parent->insertChild( this );
+	    parent->setChildrenAllocatedDirty();
+	    parent->paintable_region_dirty = TRUE;
+	}
     }
     bool     enable = isEnabled();		// remember status
     FocusPolicy fp = focusPolicy();
@@ -362,9 +402,9 @@ void QWidget::reparentSys( QWidget *parent, WFlags f, const QPoint &p,
 #endif
     widget_flags = f;
     clearWState( WState_Created | WState_Visible | WState_ForceHide );
-    if ( isTopLevel() || (!parent || parent->isVisibleTo( 0 ) ) )
-	setWState( WState_ForceHide );	// new widgets do not show up in already visible parents
     create();
+    if ( isTopLevel() || (!parent || parent->isVisible() ) )
+	setWState( WState_ForceHide );	// new widgets do not show up in already visible parents
     /*
     if ( bgp )
 	XSetWindowBackgroundPixmap( dpy, winid, bgp->handle() );
@@ -398,8 +438,8 @@ QPoint QWidget::mapToGlobal( const QPoint &pos ) const
     int	   x=pos.x(), y=pos.y();
     const QWidget* w = this;
     while (w) {
-	x += w->x();
-	y += w->y();
+	x += w->crect.x();
+	y += w->crect.y();
 	w = w->isTopLevel() ? 0 : w->parentWidget();
     }
     return QPoint( x, y );
@@ -410,23 +450,23 @@ QPoint QWidget::mapFromGlobal( const QPoint &pos ) const
     int	   x=pos.x(), y=pos.y();
     const QWidget* w = this;
     while (w) {
-	x -= w->x();
-	y -= w->y();
+	x -= w->crect.x();
+	y -= w->crect.y();
 	w = w->isTopLevel() ? 0 : w->parentWidget();
     }
     return QPoint( x, y );
 }
 
 void QWidget::setMicroFocusHint( int x, int y, int width, int height,
-				 bool /*text*/, QFont *)
+				 bool text, QFont *)
 {
     if ( QRect( x, y, width, height ) != microFocusHint() )
 	extraData()->micro_focus_hint.setRect( x, y, width, height );
-
-//XXX not implemented
-#if 0
+#ifndef QT_NO_QWS_IM
     if ( text ) {
-
+	QPoint p( x, y + height );
+	QPoint gp = mapToGlobal( p );
+	qwsDisplay()->setMicroFocus( gp.x(), gp.y());
     }
 #endif
 }
@@ -490,26 +530,23 @@ void QWidget::setBackgroundEmpty()
 
 void QWidget::setCursor( const QCursor &cursor )
 {
-    if ( 1/*cursor.handle() != arrowCursor.handle()*/
-	 || (extra && extra->curs) ) {
-	createExtra();
-	delete extra->curs;
-	extra->curs = new QCursor(cursor);
-    }
+    createExtra();
+    delete extra->curs;
+    extra->curs = new QCursor(cursor);
     setWState( WState_OwnCursor );
-//    QCursor *oc = QApplication::overrideCursor();
+    if ( isVisible() )
+	updateCursor( paintableRegion() );
 }
 
 void QWidget::unsetCursor()
 {
-    if ( !isTopLevel() ) {
-	if (extra ) {
-	    delete extra->curs;
-	    extra->curs = 0;
-	}
-	clearWState( WState_OwnCursor );
-	// XXX XDefineCursor( x11Display(), winId(), None );
+    if ( extra ) {
+	delete extra->curs;
+	extra->curs = 0;
     }
+    clearWState( WState_OwnCursor );
+    if ( isVisible() )
+	updateCursor( paintableRegion() );
 }
 #endif //QT_NO_CURSOR
 
@@ -576,7 +613,7 @@ void QWidget::grabMouse( const QCursor &cursor )
 	mouseGrb->releaseMouse();
 
     qwsDisplay()->grabMouse(this,TRUE);
-    qwsDisplay()->selectCursor(this, (int)cursor.handle());
+    qwsDisplay()->selectCursor(this, (unsigned int)cursor.handle());
     mouseGrb = this;
 }
 #endif
@@ -674,7 +711,7 @@ void QWidget::repaint( const QRegion& reg, bool erase )
     if ( (widget_state & (WState_Visible|WState_BlockUpdates)) == WState_Visible && isVisible() ) {
 	if ( erase )
 	    this->erase(reg);
-	QPaintEvent e( reg );
+	QPaintEvent e( reg, erase );
 	qt_set_paintevent_clipping( this, reg );
 	QApplication::sendEvent( this, &e );
 	qt_clear_paintevent_clipping();
@@ -684,6 +721,7 @@ void QWidget::repaint( const QRegion& reg, bool erase )
 void QWidget::showWindow()
 {
     if ( testWFlags(WType_TopLevel) ) {
+	updateRequestedRegion( mapToGlobal(QPoint(0,0)) );
 	QRegion r( req_region );
 #ifndef QT_NO_QWS_MANAGER
 	if ( extra && extra->topextra && extra->topextra->qwsManager ) {
@@ -697,15 +735,15 @@ void QWidget::showWindow()
 	    qwsDisplay()->requestFocus(winId(),TRUE);
 	}
 	qwsDisplay()->setAltitude( winId(),
-	    topData()->fullscreen ? 2 :
 		testWFlags(WStyle_StaysOnTop) ? 1 : 0, TRUE );
 
     } else if ( !topLevelWidget()->in_show ) {
+	updateRequestedRegion( mapToGlobal(QPoint(0,0)) );
 	QWidget *p = parentWidget();
-	p->setChildrenAllocatedDirty();
+	p->setChildrenAllocatedDirty( geometry(), this );
 	p->paintable_region_dirty = TRUE;
 	p->overlapping_children = -1;
-	update();
+	paint_heirarchy( this, TRUE );
     }
 }
 
@@ -720,15 +758,18 @@ void QWidget::hideWindow()
 	qwsDisplay()->requestFocus(winId(),FALSE);
     } else {
 	QWidget *p = parentWidget();
-	p->setChildrenAllocatedDirty();
-	p->paintable_region_dirty = TRUE;
-	if ( p->overlapping_children )
-	    p->overlapping_children = -1;
-	if ( p->isVisible() ) {
-	    QApplication::postEvent( p, new QPaintEvent(geometry(), TRUE) );
-	    paint_children( p,geometry(),TRUE );
+	if ( p ) {
+	    p->setChildrenAllocatedDirty( geometry(), this );
+	    p->paintable_region_dirty = TRUE;
+	    if ( p->overlapping_children )
+		p->overlapping_children = -1;
+	    if ( p->isVisible() ) {
+		QApplication::postEvent( p, new QPaintEvent(geometry(), TRUE) );
+		paint_children( p,geometry(),TRUE );
+	    }
 	}
     }
+    updateRequestedRegion( mapToGlobal(QPoint(0,0)) );
 }
 
 void QWidget::showMinimized()
@@ -755,38 +796,38 @@ bool QWidget::isMinimized() const
 void QWidget::showMaximized()
 {
     in_show_maximized = 1;
+    clearWState( WState_Minimized );
+    setWState(WState_Maximized);
     if ( testWFlags(WType_TopLevel) ) {
 	createTLExtra();
+	if ( topData()->normalGeometry.width() < 0 )
+	    topData()->normalGeometry = geometry();
 #ifndef QT_NO_QWS_MANAGER
-	if ( extra && extra->topextra && extra->topextra->qwsManager ) {
+	if ( extra && extra->topextra && extra->topextra->qwsManager )
 	    extra->topextra->qwsManager->maximize();
-	} else
+	else
 #endif
-	{
 	    setGeometry( qt_maxWindowRect );
-	}
     }
     show();
     QEvent e( QEvent::ShowMaximized );
     QApplication::sendEvent( this, &e );
-    clearWState( WState_Minimized );
-    setWState(WState_Maximized);
     in_show_maximized = 0;
 }
 
 void QWidget::showNormal()
 {
     if ( isTopLevel() ) {
-	if ( topData()->fullscreen ) {
+	if ( topData()->fullscreen )
 	    reparent( 0, WType_TopLevel, QPoint(0,0) );
-	    topData()->fullscreen = 0;
-	}
 	QRect r = topData()->normalGeometry;
 	if ( r.width() >= 0 ) {
 	    topData()->normalGeometry = QRect(0,0,-1,-1);
 	    setGeometry( r );
 	}
     }
+    if ( extra && extra->topextra )
+	extra->topextra->fullscreen = 0;
     show();
     clearWState( WState_Minimized | WState_Maximized );
 }
@@ -802,8 +843,8 @@ void QWidget::raise()
 	    setActiveWindow();
 	qwsDisplay()->setAltitude( winId(), 0 );
     } else if ( p ) {
-	p->setChildrenAllocatedDirty();
-	paint_children( p,geometry(),TRUE );
+	p->setChildrenAllocatedDirty( geometry(), this );
+	paint_heirarchy( this, TRUE );
     }
 }
 
@@ -815,7 +856,7 @@ void QWidget::lower()
     if ( isTopLevel() )
 	qwsDisplay()->setAltitude( winId(), -1 );
     else if ( p ) {
-	p->setChildrenAllocatedDirty();
+	p->setChildrenAllocatedDirty( geometry() );
 	paint_children( p,geometry(),TRUE );
     }
 }
@@ -867,10 +908,10 @@ void QWidget::internalSetGeometry( int x, int y, int w, int h, bool isMove )
 	w = 1;
     if ( h < 1 )
 	h = 1;
-    QPoint oldp = pos();
-    QSize  olds = size();
 
-    QRect  r( x, y, w, h );
+    QPoint oldp = geometry().topLeft();
+    QSize olds = size();
+    QRect r( x, y, w, h );
 
     bool isResize = olds != r.size();
 
@@ -884,16 +925,16 @@ void QWidget::internalSetGeometry( int x, int y, int w, int h, bool isMove )
 	oldAlloc = allocatedRegion();
     }
 
-    if ( !r.contains(geometry()) ) // eg. might not FIT in max window rect
+    if ( !in_show_maximized && !r.contains(geometry()) ) // eg. might not FIT in max window rect
 	clearWState(WState_Maximized);
 
+    QPoint oldPos = pos();
     crect = r;
 
     if ( testWFlags(WType_Desktop) )
 	return;
 
     if ( isTopLevel() ) {
-
 	//### ConfigPending not implemented, do we need it?
 	//setWState( WState_ConfigPending );
 	if ( isMove && ( w==olds.width() && h==olds.height() ) ) {
@@ -916,6 +957,7 @@ void QWidget::internalSetGeometry( int x, int y, int w, int h, bool isMove )
 	if ( isVisible() ) {
 	    if ( isMove && !isResize && alloc_region_index >= 0 ) {
 		qwsDisplay()->moveRegion(winId(), x - oldp.x(), y - oldp.y());
+		setChildrenAllocatedDirty();
 	    } else {
 		QRegion rgn( req_region );
 #ifndef QT_NO_QWS_MANAGER
@@ -926,6 +968,14 @@ void QWidget::internalSetGeometry( int x, int y, int w, int h, bool isMove )
 		}
 #endif
 		qwsDisplay()->requestRegion(winId(), rgn);
+		if ( extra && extra->topextra ) {
+		    QRect br( rgn.boundingRect() );
+		    br = qt_screen->mapFromDevice( br, QSize(qt_screen->deviceWidth(), qt_screen->deviceHeight()) );
+		    extra->topextra->fleft = crect.x()-br.x();
+		    extra->topextra->ftop = crect.y()-br.y();
+		    extra->topextra->fright = br.right()-crect.right();
+		    extra->topextra->fbottom = br.bottom()-crect.bottom();
+		}
 	    }
 	}
     }
@@ -933,7 +983,7 @@ void QWidget::internalSetGeometry( int x, int y, int w, int h, bool isMove )
     if ( isVisible() ) {
 	isSettingGeometry = TRUE;
 	if ( isMove ) {
-	    QMoveEvent e( r.topLeft(), oldp );
+	    QMoveEvent e( pos(), oldPos );
 	    QApplication::sendEvent( this, &e );
 #ifndef QT_NO_QWS_MANAGER
 	    if (extra && extra->topextra && extra->topextra->qwsManager)
@@ -957,18 +1007,18 @@ void QWidget::internalSetGeometry( int x, int y, int w, int h, bool isMove )
 */
 	}
 
+	updateRequestedRegion( mapToGlobal(QPoint(0,0)) );
+
 	QWidget *p = parentWidget();
 	if ( !isTopLevel() || isResize ) {
 	    if ( p && !isTopLevel() ) {
 		p->paintable_region_dirty = TRUE;
 		QRegion oldr( QRect(oldp, olds) );
-		QRegion upd( ( QRegion(r) | oldr ) & p->rect() );
+		dirtyChildren.translate( x, y );
 		if ( p->isSettingGeometry ) {
-		    dirtyChildren.translate( x, y );
 		    if ( oldp != r.topLeft() ) {
+			QRegion upd( (QRegion(r) | oldr) & p->rect() );
 			dirtyChildren |= upd;
-			QApplication::postEvent( this, new QPaintEvent(rect(),
-			    !testWFlags(QWidget::WRepaintNoErase)) );
 		    } else {
 			dirtyChildren |= QRegion(r) - oldr;
 			QApplication::postEvent( this, new QPaintEvent(rect(),
@@ -976,6 +1026,9 @@ void QWidget::internalSetGeometry( int x, int y, int w, int h, bool isMove )
 		    }
 		    p->dirtyChildren |= dirtyChildren;
 		} else {
+		    QRegion upd( (QRegion(r) | oldr) & p->rect() );
+		    dirtyChildren |= upd;
+		    QRegion paintRegion = dirtyChildren;
 #define FAST_WIDGET_MOVE
 #ifdef FAST_WIDGET_MOVE
 		    if ( isMove && ( w==olds.width() && h==olds.height() ) ) {
@@ -999,25 +1052,31 @@ void QWidget::internalSetGeometry( int x, int y, int w, int h, bool isMove )
 			    scrollRegion = qt_screen->mapFromDevice( scrollRegion, ds );
 			    QPoint gp = p->mapToGlobal( QPoint(0,0) );
 			    scrollRegion.translate( -gp.x(), -gp.y() );
-			    upd -= scrollRegion;
+			    paintRegion -= scrollRegion;
 			}
 		    }
 #endif
-		    if ( !upd.isEmpty() )
-			QApplication::postEvent( p, new QPaintEvent(upd, TRUE) );
-		    dirtyChildren.translate( x, y );
-		    dirtyChildren |= upd;
-		    p->setChildrenAllocatedDirty();
-		    paint_children( p, dirtyChildren & p->rect(), FALSE );
+		    if ( !oldr.isEmpty() )
+			QApplication::postEvent( p, new QPaintEvent(oldr, TRUE) );
+		    p->setChildrenAllocatedDirty( dirtyChildren, this );
+		    updateActivePainter();
+		    paint_children( p, paintRegion, isResize );
 		}
 		p->overlapping_children = -1;
 	    } else {
-		QApplication::postEvent( this, new QPaintEvent(rect(),
-		    !testWFlags(QWidget::WRepaintNoErase)) );
-		if ( oldp != r.topLeft() )
-		    dirtyChildren |= rect();
-		paint_children( this, dirtyChildren, TRUE );
+		if ( oldp != r.topLeft() ) {
+		    updateActivePainter();
+		    paint_heirarchy( this, TRUE );
+		} else {
+		    setChildrenAllocatedDirty( dirtyChildren );
+		    updateActivePainter();
+		    QApplication::postEvent( this, new QPaintEvent(rect(),
+			!testWFlags(QWidget::WResizeNoErase)) );
+		    paint_children( this, dirtyChildren, TRUE );
+		}
 	    }
+	} else {
+	    updateActivePainter();
 	}
 #ifndef QT_NO_QWS_MANAGER
 	if (isResize && extra && extra->topextra && extra->topextra->qwsManager) {
@@ -1030,22 +1089,24 @@ void QWidget::internalSetGeometry( int x, int y, int w, int h, bool isMove )
     } else {
 	if ( isMove )
 	    QApplication::postEvent( this,
-				     new QMoveEvent( r.topLeft(), oldp ) );
+				     new QMoveEvent( pos(), oldPos ) );
 	if ( isResize )
 	    QApplication::postEvent( this,
-				     new QResizeEvent( r.size(), olds ) );
+				     new QResizeEvent(crect.size(), olds) );
     }
 }
 
 
 void QWidget::setMinimumSize( int minw, int minh )
 {
-    // This is really just a work-around. Layout shouldn't be asking
-    // for minimum sizes bigger than the screen.
-    if ( minw > qt_maxWindowRect.width() )
-	minw = qt_maxWindowRect.width();
-    if ( minh > qt_maxWindowRect.height() )
-	minh = qt_maxWindowRect.height();
+    if ( !qt_maxWindowRect.isEmpty() ) {
+	// This is really just a work-around. Layout shouldn't be asking
+	// for minimum sizes bigger than the screen.
+	if ( minw > qt_maxWindowRect.width() )
+	    minw = qt_maxWindowRect.width();
+	if ( minh > qt_maxWindowRect.height() )
+	    minh = qt_maxWindowRect.height();
+    }
 
 #if defined(QT_CHECK_RANGE)
     if ( minw < 0 || minh < 0 )
@@ -1137,22 +1198,16 @@ void QWidget::erase( const QRegion& reg )
     if ( backgroundMode() == NoBackground || testWFlags(WPaintUnclipped) )
 	return;
 
+    bool unclipped = testWFlags( WPaintUnclipped );
+    clearWFlags( WPaintUnclipped );
     QPainter p(this);
     p.setClipRegion( reg );
     if ( extra && extra->bg_pix ) {
 	if ( !extra->bg_pix->isNull() ) {
-	    int xoff = 0;
-	    int yoff = 0;
-	    if ( !isTopLevel() ) {
-		if ( backgroundOrigin() == QWidget::ParentOrigin ) {
-		    xoff = x();
-		    yoff = y();
-		} else if ( backgroundOrigin() == QWidget::WindowOrigin ) {
-		    QPoint off = mapTo( topLevelWidget(), QPoint(0,0) );
-		    xoff = off.x();
-		    yoff = off.y();
-		}
-	    }
+	    QPoint offset = backgroundOffset();
+	    int xoff = offset.x();
+	    int yoff = offset.y();
+
 	    p.drawTiledPixmap(rect(),*extra->bg_pix,
 			      QPoint(xoff%extra->bg_pix->width(),
 				     yoff%extra->bg_pix->height()));
@@ -1160,6 +1215,8 @@ void QWidget::erase( const QRegion& reg )
     } else {
 	p.fillRect(rect(),bg_col);
     }
+    if ( unclipped )
+	setWFlags( WPaintUnclipped );
 }
 
 void QWidget::scroll( int dx, int dy )
@@ -1169,7 +1226,7 @@ void QWidget::scroll( int dx, int dy )
 
 void QWidget::scroll( int dx, int dy, const QRect& r )
 {
-    if ( testWState( WState_BlockUpdates ) )
+    if ( testWState( WState_BlockUpdates ) && !children() )
 	return;
     bool valid_rect = r.isValid();
     QRect sr = valid_rect?r:rect();
@@ -1215,6 +1272,8 @@ void QWidget::scroll( int dx, int dy, const QRect& r )
 
     paintable_region_dirty = TRUE;
 
+    QPoint gpos = mapToGlobal( QPoint() );
+
     if ( !valid_rect && children() ) {	// scroll children
 	setChildrenAllocatedDirty();
 	QPoint pd( dx, dy );
@@ -1227,6 +1286,7 @@ void QWidget::scroll( int dx, int dy, const QRect& r )
 		QPoint oldp = w->pos();
 		QRect  r( w->pos() + pd, w->size() );
 		w->crect = r;
+		w->updateRequestedRegion( gpos + w->pos() );
 		QMoveEvent e( r.topLeft(), oldp );
 		QApplication::sendEvent( w, &e );
 	    }
@@ -1234,11 +1294,9 @@ void QWidget::scroll( int dx, int dy, const QRect& r )
 	}
     }
 
-    QPoint p = mapToGlobal( QPoint() );
-
     QSize ds( qt_screen->deviceWidth(), qt_screen->deviceHeight() );
     scrollRegion = qt_screen->mapFromDevice( scrollRegion, ds );
-    scrollRegion.translate( -p.x(), -p.y() );
+    scrollRegion.translate( -gpos.x(), -gpos.y() );
 
     QRegion update( sr );
     update -= scrollRegion;
@@ -1339,38 +1397,53 @@ void QWidget::updateOverlappingChildren() const
 	    ++it;
 	    if ( ch->isWidgetType() && !((QWidget*)ch)->isTopLevel() ) {
 		QWidget *w = (QWidget *)ch;
-		QRegion rr( w->requestedRegion() );
-		QRegion ir = r & rr;
-		if ( !ir.isEmpty() ) {
-		    overlapping_children = 1;
-		    return;
+		if ( w->isVisible() ) {
+		    QRegion rr( w->req_region );
+		    QRegion ir = r & rr;
+		    if ( !ir.isEmpty() ) {
+			overlapping_children = 1;
+			return;
+		    }
+		    r |= rr;
 		}
-		r |= rr;
 	    }
 	}
     }
     overlapping_children = 0;
 }
 
-QRegion QWidget::requestedRegion() const
+void QWidget::updateRequestedRegion( const QPoint &gpos )
 {
-    if ( isTopLevel() ) {
-	return req_region;
-    } else {
-	if ( testWState( WState_Visible ) ) {
-	    QPoint gpos( mapToGlobal(QPoint(0,0)) );
-	    QRegion r( QRect(gpos,crect.size()) );
+    if ( !isTopLevel() ) {
+	if ( !testWState( WState_Visible ) || testWState(WState_ForceHide) ) {
+	    req_region = QRegion();
+	} else {
+	    req_region = QRect(gpos,crect.size());
 	    if ( extra && !extra->mask.isNull() ) {
 		QRegion maskr = extra->mask;
 		maskr.translate( gpos.x(), gpos.y() );
-		r &= maskr;
+		req_region &= maskr;
 	    }
-	    r = qt_screen->mapToDevice( r, QSize(qt_screen->width(), qt_screen->height()) );
-	    return r;
-	} else {
-	    return QRegion();
+	    req_region = qt_screen->mapToDevice( req_region, QSize(qt_screen->width(), qt_screen->height()) );
 	}
     }
+    const QObjectList *c = children();
+    if ( c ) {
+	QObjectListIt it(*c);
+	QObject* ch;
+	while ((ch=it.current())) {
+	    ++it;
+	    if ( ch->isWidgetType() && !((QWidget*)ch)->isTopLevel() ) {
+		QWidget *w = (QWidget *)ch;
+		w->updateRequestedRegion( gpos + w->pos() );
+	    }
+	}
+    }
+}
+
+QRegion QWidget::requestedRegion() const
+{
+    return req_region;
 }
 
 void QWidget::setChildrenAllocatedDirty()
@@ -1388,6 +1461,25 @@ void QWidget::setChildrenAllocatedDirty()
     }
 }
 
+void QWidget::setChildrenAllocatedDirty( const QRegion &r, const QWidget *dirty )
+{
+    const QObjectList *c = children();
+    if ( c ) {
+	QObjectListIt it(*c);
+	QObject* ch;
+	while ((ch=it.current())) {
+	    ++it;
+	    if ( ch->isWidgetType() ) {
+		QWidget *w = (QWidget *)ch;
+		if ( r.boundingRect().intersects( w->geometry() ) )
+		    w->alloc_region_dirty = TRUE;
+		if ( w == dirty )
+		    break;
+	    }
+	}
+    }
+}
+
 // check my hierarchy for dirty allocated regions
 bool QWidget::isAllocatedRegionDirty() const
 {
@@ -1414,7 +1506,7 @@ QRegion QWidget::allocatedRegion() const
 	} else {
 	    if ( isAllocatedRegionDirty() ) {
 		const QObjectList *c;
-		QRegion r( requestedRegion() );
+		QRegion r( req_region );
 		r &= parentWidget()->allocatedRegion();
 		parentWidget()->updateOverlappingChildren();
 		if ( parentWidget()->overlapping_children ) {
@@ -1426,11 +1518,12 @@ QRegion QWidget::allocatedRegion() const
 			while ((ch=it.current())) {
 			    ++it;
 			    if ( ch->isWidgetType() ) {
-				if ( ((QWidget*)ch) == this )
+				QWidget *w = (QWidget*)ch;
+				if ( w == this )
 				    clip=TRUE;
-				else if ( clip && !((QWidget*)ch)->isTopLevel()) {
-				    if ( ((QWidget*)ch)->geometry().intersects( geometry() ) )
-					r -= ((QWidget*)ch)->requestedRegion();
+				else if ( clip && !w->isTopLevel() && w->isVisible() ) {
+				    if ( w->geometry().intersects( geometry() ) )
+					r -= w->req_region;
 				}
 			    }
 			}
@@ -1472,12 +1565,17 @@ QRegion QWidget::paintableRegion() const
 		QObject* ch;
 		while ((ch=it.current())) {
 		    ++it;
-		    if ( ch->isWidgetType() && !((QWidget*)ch)->isTopLevel() ) {
-			paintable_region -= ((QWidget*)ch)->requestedRegion();
+		    if ( ch->isWidgetType() && !((QWidget*)ch)->isTopLevel() && ((QWidget*)ch)->isVisible() ) {
+			paintable_region -= ((QWidget*)ch)->req_region;
 		    }
 		}
 	    }
 	    paintable_region_dirty = FALSE;
+#ifndef QT_NO_CURSOR
+	    // The change in paintable region may have result in the
+	    // cursor now being within my region.
+	    updateCursor( paintable_region );
+#endif
 	}
 	if ( !isTopLevel() )
 	    return paintable_region;
@@ -1494,6 +1592,52 @@ QRegion QWidget::paintableRegion() const
     return QRegion();
 }
 
+static QPtrDict<QPainter> *painterDict = 0;
+
+void QWidget::setActivePainter( QPainter *painter ) const
+{
+    if ( !painterDict )
+	painterDict = new QPtrDict<QPainter>;
+    painterDict->insert( (void *)this, painter );
+}
+
+void QWidget::clearActivePainter() const
+{
+    if ( painterDict )
+	painterDict->remove( (void *)this );
+}
+
+class QWSUpdatePainter : public QPainter
+{
+    public:
+	void setGfx( QGfx *g ) {
+	    gfx = g;
+	}
+};
+
+void QWidget::updateActivePainter() const
+{
+    if ( painterDict && painterDict->count() ) {
+	const QObjectList *c = children();
+	if ( c ) {
+	    QObjectListIt it(*c);
+	    QObject* ch;
+	    while ((ch=it.current())) {
+		++it;
+		if ( ch->isWidgetType() ) {
+		    QWidget *w = (QWidget *)ch;
+		    w->updateActivePainter();
+		}
+	    }
+	}
+	QPainter *painter = painterDict->find( (void *)this );
+	if ( painter ) {
+	    painter->save();
+	    ((QWSUpdatePainter *)painter)->setGfx( graphicsContext() );
+	    painter->restore();
+	}
+    }
+}
 
 void QWidget::setMask( const QRegion& region )
 {
@@ -1527,6 +1671,7 @@ void QWidget::setMask( const QRegion& region )
 #endif
 	    qwsDisplay()->requestRegion(winId(), rgn);
 	} else {
+	    updateRequestedRegion( mapToGlobal(QPoint(0,0)) );
 	    parentWidget()->paintable_region_dirty = TRUE;
 	    parentWidget()->repaint(geometry());
 	    paint_children( parentWidget(),geometry(),TRUE );
@@ -1559,6 +1704,13 @@ QGfx * QWidget::graphicsContext(bool clip_children) const
 {
     QGfx * qgfx_qws;
     qgfx_qws=qwsDisplay()->screenGfx();
+    updateGraphicsContext( qgfx_qws, clip_children );
+
+    return qgfx_qws;
+}
+
+void QWidget::updateGraphicsContext( QGfx *qgfx_qws, bool clip_children ) const
+{
     QPoint offset=mapToGlobal(QPoint(0,0));
     QRegion r; // empty if not visible
     if ( isVisible() && topLevelWidget()->isVisible() ) {
@@ -1593,8 +1745,6 @@ QGfx * QWidget::graphicsContext(bool clip_children) const
     if ( extra && extra->topextra && extra->topextra->qwsManager )
 	qgfx_qws->setClipRegion(rect());
 #endif
-
-    return qgfx_qws;
 }
 
 /*!
@@ -1622,6 +1772,13 @@ bool QWidget::isMaximized() const
     return testWState(WState_Maximized);
 }
 
+void QWidget::resetInputContext()
+{
+#ifndef QT_NO_QWS_IM
+    QInputContext::reset();
+#endif
+}
+
 void QWidget::updateFrameStrut() const
 {
     QWidget *that = (QWidget *) this;
@@ -1634,7 +1791,16 @@ void QWidget::updateFrameStrut() const
     //FIXME: need to fill in frame strut info
 }
 
-// This one
-void QWidget::resetInputContext()
+#ifndef QT_NO_CURSOR
+void QWidget::updateCursor( const QRegion &r ) const
 {
+    if ( (!QWidget::mouseGrabber() || QWidget::mouseGrabber() == this) &&
+	    qt_last_cursor != (WId)cursor().handle() ) {
+	QSize s( qt_screen->width(), qt_screen->height() );
+	QPoint pos = qt_screen->mapToDevice(QPoint(qt_last_x, qt_last_y), s);
+	if ( r.contains(pos) )
+	    qwsDisplay()->selectCursor((QWidget*)this, (unsigned int)cursor().handle());
+    }
 }
+#endif
+
