@@ -1,5 +1,5 @@
 /**********************************************************************
-** $Id:  qt/qgroupbox.cpp   3.0.3   edited Dec 7 16:50 $
+** $Id: qt/qgroupbox.cpp   3.1.2   edited Jan 6 18:59 $
 **
 ** Implementation of QGroupBox widget class
 **
@@ -79,7 +79,9 @@
   constructor) and the title's alignment().
 
   You can change the spacing used by the group box with
-  setInsideMargin() and setInsideSpacing().
+  setInsideMargin() and setInsideSpacing(). To reduce space consumption,
+  you can remove the right, left and bottom edges of the frame with
+  setFlat().
 
   <img src=qgrpbox-w.png>
 
@@ -150,10 +152,8 @@ QGroupBox::QGroupBox( int strips, Orientation orientation,
 
 void QGroupBox::init()
 {
-    int fs;
     align = AlignAuto;
-    fs = QFrame::Box | QFrame::Sunken;
-    setFrameStyle( fs );
+    setFrameStyle( QFrame::GroupBoxPanel | QFrame::Sunken );
 #ifndef QT_NO_ACCEL
     accel = 0;
 #endif
@@ -165,34 +165,37 @@ void QGroupBox::init()
     dir = Horizontal;
     marg = 11;
     spac = 5;
+    bFlat = FALSE;
 }
 
 void QGroupBox::setTextSpacer()
 {
-    QSpacerItem *sp = (QSpacerItem*)d;
-    if ( ! sp )
+    QSpacerItem *spacer = (QSpacerItem*)d;
+    if ( !spacer )
 	return;
     int h = 0;
     int w = 0;
     if ( lenvisible ) {
 	QFontMetrics fm = fontMetrics();
-	h = fm.height();
+	int fh = fm.height();
 	w = fm.width( str, lenvisible ) + 2*fm.width( "xx" );
+	h = frameRect().y();
 	if ( layout() ) {
-	    int m = layout()->margin();
+ 	    int m = layout()->margin();
+	    int sp = layout()->spacing();
 	    // do we have a child layout?
 	    for ( QLayoutIterator it = layout()->iterator(); it.current(); ++it ) {
 		if ( it.current()->layout() ) {
-		    m += it.current()->layout()->margin();
+ 		    m += it.current()->layout()->margin();
+		    sp = QMAX( sp, it.current()->layout()->spacing() );
 		    break;
 		}
 	    }
-	    if ( m > 4 )
-		h -= m - 4;
-	    h = QMAX( 0, h );
+	    h = QMAX( fh-m, h );
+	    h += QMAX( sp - (h+m - fh), 0 );
 	}
     }
-    sp->changeSize( w, h, QSizePolicy::Minimum, QSizePolicy::Fixed );
+    spacer->changeSize( w, h, QSizePolicy::Minimum, QSizePolicy::Fixed );
 }
 
 
@@ -214,13 +217,6 @@ void QGroupBox::setTitle( const QString &title )
 #endif
     calculateFrame();
     setTextSpacer();
-    if ( layout() ) {
-	layout()->activate();
-	QSize s( size() );
-	QSize ms( minimumSizeHint() );
-	resize( QMAX( s.width(), ms.width() ),
-		QMAX( s.height(), ms.height() ) );
-    }
 
     update();
     updateGeometry();
@@ -293,31 +289,40 @@ void QGroupBox::paintEvent( QPaintEvent *event )
     if ( lenvisible ) {					// draw title
 	QFontMetrics fm = paint.fontMetrics();
 	int h = fm.height();
-	int tw = fm.width( str, lenvisible ) + 2*fm.width(QChar(' '));
+	int tw = fm.width( str, lenvisible ) + fm.width(QChar(' '));
 	int x;
+	int marg = bFlat ? 0 : 8;
 	if ( align & AlignHCenter )		// center alignment
 	    x = frameRect().width()/2 - tw/2;
 	else if ( align & AlignRight )	// right alignment
-	    x = frameRect().width() - tw - 8;
+	    x = frameRect().width() - tw - marg;
 	else if ( align & AlignLeft )		 // left alignment
-	    x = 8;
+	    x = marg;
 	else { // auto align
 	    if( QApplication::reverseLayout() )
-		x = frameRect().width() - tw - 8;
+		x = frameRect().width() - tw - marg;
 	    else
-		x = 8;
+		x = marg;
 	}
 	QRect r( x, 0, tw, h );
-
-	style().drawItem( &paint, r, AlignCenter | ShowPrefix, colorGroup(),
-			  isEnabled(), 0, str );
+	int va = style().styleHint(QStyle::SH_GroupBox_TextLabelVerticalAlignment, this);
+	if(va & AlignTop)
+	    r.moveBy(0, fm.descent());
+	QColor pen( (QRgb) style().styleHint(QStyle::SH_GroupBox_TextLabelColor, this )  );
+	style().drawItem( &paint, r, ShowPrefix | AlignHCenter | va, colorGroup(),
+			  isEnabled(), 0, str, -1, ownPalette() ? 0 : &pen );
 	paint.setClipRegion( event->region().subtract( r ) ); // clip everything but title
     }
-    // I really think this should call drawFrame() instead...
-    style().drawPrimitive( QStyle::PE_GroupBoxFrame, &paint, frameRect(),
-			   colorGroup(), QStyle::Style_Default,
-			   QStyleOption(lineWidth(), midLineWidth(),
-			    frameShape(), frameShadow()) );
+    if ( bFlat ) {
+	    QRect fr = frameRect();
+	    QPoint p1( fr.x(), fr.y() + 1 );
+            QPoint p2( fr.x() + fr.width(), p1.y() );
+	    // ### This should probably be a style primitive.
+            qDrawShadeLine( &paint, p1, p2, colorGroup(), TRUE,
+                            lineWidth(), midLineWidth() );
+    } else {
+	drawFrame(&paint);
+    }
     drawContents( &paint );			// draw the contents
 }
 
@@ -358,7 +363,7 @@ void QGroupBox::addSpace( int size )
 
 /*!
   \property QGroupBox::columns
-  \brief the number of columns or rows (depending on \l orientation) in the group box
+  \brief the number of columns or rows (depending on \l QGroupBox::orientation) in the group box
 
   Usually it is not a good idea to set this property because it is slow
   (it does a complete layout).  It is better to set the number of columns
@@ -611,16 +616,19 @@ void QGroupBox::calculateFrame()
 
     if ( lenvisible ) { // do we have a label?
 	QFontMetrics fm = fontMetrics();
-	int h = fm.height();
 	while ( lenvisible ) {
-	    int tw = fm.width( str, lenvisible ) + 2*fm.width(QChar(' '));
+	    int tw = fm.width( str, lenvisible ) + 4*fm.width(QChar(' '));
 	    if ( tw < width() )
 		break;
 	    lenvisible--;
 	}
 	if ( lenvisible ) { // but do we also have a visible label?
 	    QRect r = rect();
-	    r.setTop( h/2 );				// frame rect should be
+	    int va = style().styleHint(QStyle::SH_GroupBox_TextLabelVerticalAlignment, this);
+	    if(va & AlignVCenter)
+		r.setTop( fm.height()/2 );				// frame rect should be
+	    else if(va & AlignTop)
+		r.setTop(fm.ascent());
 	    setFrameRect( r );			//   smaller than client rect
 	    return;
 	}
@@ -673,4 +681,27 @@ QSize QGroupBox::sizeHint() const
     }
 }
 
+/*!
+  \property QGroupBox::flat
+  \brief whether the group box is painted flat or has a frame around it.
+
+  By default a group box has a surrounding frame, with the title being placed
+  on the upper frame line. In flat mode the right, left and bottom frame lines
+  are omitted, and only the thin line at the top is drawn.
+
+  \sa title
+*/
+bool QGroupBox::isFlat() const
+{
+    return bFlat;
+}
+
+void QGroupBox::setFlat( bool b )
+{
+    if ( (bool)bFlat == b )
+	return;
+    bFlat = b;
+    update();
+}
+
 #endif
